# ClusterCardList.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Plugin to find clusters in the card lists."""

import gtk
import random
import math
from sutekh.core.SutekhObjects import PhysicalCard, \
                                      PhysicalCardSet, \
                                      IPhysicalCard
from sutekh.core.CardListTabulator import CardListTabulator
from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow
from sutekh.gui.SutekhDialog import SutekhDialog, do_complaint_error


class ClusterCardList(SutekhPlugin):
    """Plugin that attempts to find clusters in the card list.

       Allows the user to choose various clustering parameters, such as
       attributes to to use and methods, and then to create card sets
       from the clustering results."""
    # pylint: disable-msg=R0902
    # we use a lot of attributes to keep track of data & gui aspects

    dTableVersions = {}
    aModelsSupported = (PhysicalCard, PhysicalCardSet)

    # pylint: disable-msg=W0142
    # ** magic OK
    def __init__(self, *args, **kwargs):
        super(ClusterCardList, self).__init__(*args, **kwargs)
        # pylint: disable-msg=C0103
        # fMakeCardFromCluster triggers on length, but we like the name
        if not self.model:
            self._fMakeCardSetFromCluster = None
        self._fMakeCardSetFromCluster = self._make_pcs_from_cluster
        self._dPropButtons = {}
        self._dGroups = {}
        self._oResultsVbox = None
        self._oNotebook = None
        self._aDistanceMeasureGroup = None

        # cluster parameter widgets
        self._oAutoNumClusters = None
        self._oNumClustersSpin = None
        self._oNumIterSpin = None

    # pylint: enable-msg=W0142

    def get_menu_item(self):
        """Register on the 'Analyze' menu."""
        if not self.check_versions() or not self.check_model_type():
            return None
        oCluster = gtk.MenuItem("Cluster Cards")
        oCluster.connect("activate", self.activate)
        return ('Analyze', oCluster)

    def activate(self, _oWidget):
        """In response to the menu, create the correct dialog."""
        oDlg = self.make_dialog()
        oDlg.run()

    def make_dialog(self):
        """Create the dialog for the clustering options."""
        sDlgName = "Cluster Cards in List"

        self._make_prop_groups()

        oDlg = SutekhDialog(sDlgName, self.parent,
                gtk.DIALOG_DESTROY_WITH_PARENT)
        oDlg.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)
        oDlg.add_button(gtk.STOCK_EXECUTE, gtk.RESPONSE_APPLY)

        oDlg.connect("response", self.handle_response)

        self._oNotebook = gtk.Notebook()

        oTableSection = self._make_table_section()
        oAlgorithmSection = self._make_algorithm_section()
        oResultsSection = self._make_results_section()

        self._oNotebook.append_page(oTableSection, gtk.Label('Select Columns'))
        self._oNotebook.append_page(oAlgorithmSection, gtk.Label('Settings'))
        self._oNotebook.append_page(oResultsSection, gtk.Label('Results'))

        # pylint: disable-msg=E1101
        # vbox confuses pylint
        oDlg.vbox.pack_start(self._oNotebook)
        oDlg.show_all()

        return oDlg

    def _make_prop_groups(self):
        """Extract the list of possible properties to cluster on."""
        dPropFuncs = CardListTabulator.get_default_prop_funcs()
        self._dGroups = {}

        for sName, fProp in dPropFuncs.iteritems():
            aParts = sName.split(":")
            if len(aParts) == 1:
                self._dGroups.setdefault("Miscellaneous",
                        {})[sName.capitalize()] = fProp
            else:
                sGroup, sRest = aParts[0].strip().capitalize(), \
                        ":".join(aParts[1:]).strip().capitalize()
                self._dGroups.setdefault(sGroup, {})[sRest] = fProp

    def _make_table_section(self):
        """Create a notebook, and populate the first tabe with a list of
           properties to cluster on."""
        oNotebook = gtk.Notebook()
        aGroups = self._dGroups.keys()
        aGroups.sort()

        self._dPropButtons = {}

        for sGroup in aGroups:
            self._dPropButtons[sGroup] = {}
            dPropFuncs = self._dGroups[sGroup]

            oHbx = gtk.HBox(False, 0)

            iCols = 3
            iPropsPerCol = len(dPropFuncs) / iCols
            if len(dPropFuncs) % iCols != 0:
                iPropsPerCol += 1

            aPropNames = dPropFuncs.keys()
            aPropNames.sort()

            for iProp, sName in enumerate(aPropNames):
                if iProp % iPropsPerCol == 0:
                    oVbx = gtk.VBox(False, 0)
                    oHbx.pack_start(oVbx)

                oBut = gtk.CheckButton(sName)
                oVbx.pack_start(oBut, False)  # pack at top, don't fill space

                self._dPropButtons[sGroup][sName] = oBut

            oNotebook.append_page(oHbx, gtk.Label(sGroup))

        return oNotebook

    def _make_algorithm_section(self):
        """Populate a tab on the notebook with the parameters and options for
           the clustering algorithm.
           """
        oVbx = gtk.VBox(False, 0)

        # K-Means is the only clustering solution for the moment
        oHeading = gtk.Label()
        oHeading.set_markup("<b>Parameters for K-Means Clustering</b>")
        oVbx.pack_start(oHeading, False, False, 5)  # top align

        # Number of iterations
        oNumIterLabel = gtk.Label("Number of Iterations:")
        self._oNumIterSpin = gtk.SpinButton()
        self._oNumIterSpin.set_range(1, 50)
        self._oNumIterSpin.set_increments(1, 10)
        self._oNumIterSpin.set_value(10)
        oHbox = gtk.HBox(False, 0)
        oHbox.pack_start(oNumIterLabel, False)  # left align
        oHbox.pack_end(self._oNumIterSpin, False)  # right align
        oVbx.pack_start(oHbox, False)

        # Autoset Num clusters
        self._oAutoNumClusters = gtk.CheckButton("One cluster per 80 cards")
        oHbox = gtk.HBox(False, 0)
        oHbox.pack_start(self._oAutoNumClusters, False)  # right align
        oVbx.pack_start(oHbox, False)

        # Number of clusters
        oNumClustersLabel = gtk.Label("Number of Clusters:")
        self._oNumClustersSpin = gtk.SpinButton()
        self._oNumClustersSpin.set_range(2, 50)
        self._oNumClustersSpin.set_increments(1, 10)
        self._oNumClustersSpin.set_value(10)
        oHbox = gtk.HBox(False, 0)
        oHbox.pack_start(oNumClustersLabel, False)  # left align
        oHbox.pack_end(self._oNumClustersSpin, False)  # right align
        oVbx.pack_start(oHbox, False)

        # Connect Autoset and Number of clusters
        self._oAutoNumClusters.set_active(True)
        self._oNumClustersSpin.set_sensitive(False)

        def auto_toggled(oAutoNumClusters):
            """Called when oAutoNumClusters is toggled."""
            if oAutoNumClusters.get_active():
                self._oNumClustersSpin.set_sensitive(False)
            else:
                self._oNumClustersSpin.set_sensitive(True)

        self._oAutoNumClusters.connect("toggled", auto_toggled)

        # Separator
        oVbx.pack_start(gtk.HSeparator(), False, False, 10)

        # Distance Measure
        oDistLabel = gtk.Label()
        oDistLabel.set_markup("<b>Distance Measure for Clustering</b>")
        oVbx.pack_start(oDistLabel, False, False, 5)

        oIter = Vector.METRICS.iterkeys()
        for sName in oIter:
            oFirstBut = gtk.RadioButton(None, sName, False)
            oVbx.pack_start(oFirstBut, False)
            break

        for sName in oIter:
            oBut = gtk.RadioButton(oFirstBut, sName)
            oVbx.pack_start(oBut, False)

        self._aDistanceMeasureGroup = oFirstBut.get_group()

        return oVbx

    def _make_results_section(self):
        """Create a widget in the notebook which can be used to display
           the clustering results."""
        oVbx = gtk.VBox(False, 0)
        self._oResultsVbox = gtk.VBox(False, 0)

        # Results Heading
        oHeading = gtk.Label()
        oHeading.set_markup("<b>Results</b>")
        oVbx.pack_start(oHeading, False)  # top align

        # No Results Yet
        oLabel = gtk.Label("No results yet.")
        self._oResultsVbox.pack_start(oLabel)
        oVbx.pack_start(self._oResultsVbox)

        return oVbx

    # pylint: disable-msg=W0201, R0914
    # W0201: We create a lot of attributes here, which is OK, because of
    # plugin structure
    # R0914: We use lots of local variables for clarity
    def _populate_results(self, aCards, aColNames, aMeans, aClusters):
        """Populate the results tab of the notebook with the results of
           clustering run."""
        self._aCards, self._aColNames, self._aMeans, self._aClusters = \
                aCards, aColNames, aMeans, aClusters

        # clean results box
        for oChild in self._oResultsVbox.get_children():
            self._oResultsVbox.remove(oChild)

        # Setup Table
        iHeaderRows = 1
        iExtraCols = 4

        oTable = gtk.Table(rows=len(aMeans) + iHeaderRows,
                columns=iExtraCols)
        oTable.set_row_spacings(0)

        # Headings
        oLabel = gtk.Label()
        oLabel.set_markup('<b>Save Deck</b>')
        oTable.attach(oLabel, 0, 1, 0, 1, xpadding=3)

        oLabel = gtk.Label()
        oLabel.set_markup('<b>Cluster Id</b>')
        oTable.attach(oLabel, 1, 2, 0, 1, xpadding=3)

        oLabel = gtk.Label()
        oLabel.set_markup('<b>No. Cards</b>')
        oTable.attach(oLabel, 2, 3, 0, 1, xpadding=3)

        oLabel = gtk.Label()
        oLabel.set_markup('<b>Cluster Center</b>')
        oTable.attach(oLabel, 3, 4, 0, 1, xpadding=3)

        # Data
        self._dCardSetMakingButtons = {}
        for iId, oMean in enumerate(self._aMeans):
            iRow = iId + iHeaderRows
            oBut = gtk.CheckButton()
            self._dCardSetMakingButtons[iId] = oBut
            sCenterText = "\n".join([
                "%s: %.2f" % (sColName, fColVal) for sColName, fColVal
                in zip(aColNames, oMean)
                if abs(fColVal) > 0.01
            ])
            oCenterLabel = gtk.Label(sCenterText)
            oCenterLabel.set_alignment(0.0, 0.5)

            oTable.attach(oBut, 0, 1, iRow, iRow + 1)
            oTable.attach(gtk.Label(str(iId)), 1, 2, iRow, iRow + 1)
            oTable.attach(gtk.Label(str(len(aClusters[iId]))),
                2, 3, iRow, iRow + 1)
            oTable.attach(oCenterLabel, 3, 4, iRow, iRow + 1)

        # top align, using viewport to scroll
        self._oResultsVbox.pack_start(AutoScrolledWindow(oTable, True))

        oMakeCardSetsButton = gtk.Button("Make Card Sets from Selected"
                " Clusters")
        oMakeCardSetsButton.connect("clicked", self.handle_make_card_sets)
        self._oResultsVbox.pack_end(oMakeCardSetsButton, False)  # bottom align

        self._oResultsVbox.show_all()

    # pylint: enable-msg=W0201, R0914

    # Actions

    def handle_response(self, oDlg, oResponse):
        """Handle the user's response to the options dialog."""
        if oResponse == gtk.RESPONSE_CLOSE:
            oDlg.destroy()
        elif oResponse == gtk.RESPONSE_APPLY:
            self.do_clustering()
            # change to results section
            self._oNotebook.set_current_page(2)

    def handle_make_card_sets(self, _oSomeObj):
        """Create card a suitable card set from the chosen clusters"""
        for iId, oBut in self._dCardSetMakingButtons.iteritems():
            if oBut.get_active():
                self._fMakeCardSetFromCluster(iId)

    @staticmethod
    def k_means_plus_plus(aCards, iNumClust, fDist):
        """Find a set of initial centers using the k-means++ algorithm.

           See http://www.stanford.edu/~darthur/kMeansPlusPlus.pdf.
           """
        aMeans = [random.choice(aCards)]

        while len(aMeans) < iNumClust:
            aDists = []
            for oVec in aCards:
                fMinD = min((fDist(oVec, oMean) for oMean in aMeans)) ** 2
                aDists.append(fMinD)

            fSumSq = sum(aDists)
            fPick = random.uniform(0, fSumSq)

            for iCard, fMinD in enumerate(aDists):
                fPick -= fMinD
                if fPick <= 0:
                    break

            # iCard is defined because k_means doesn't call
            # this unless aCards is non-empty
            # pylint: disable-msg=W0631

            if iCard == len(aCards):
                # guard against slight possibility of being very close
                # to end of fSumSq.
                iCard -= 1

            aMeans.append(aCards[iCard])

        return aMeans

    def k_means(self, aCards, iNumClust, iIterations, fDist):
        """Perform k-means clustering on a list of cards using Lloyd's
           algorithm."""
        if (not aCards) or (not aCards[0]):
            # empty card set or zero-length vectors
            return [], []

        aCards = [Vector(x) for x in aCards]
        aMeans = self.k_means_plus_plus(aCards, iNumClust, fDist)
        iCards = len(aCards)

        # just do a fixed number of interations (no complex stopping condition)
        for _iIter in range(iIterations):
            # empty clusters
            aClusters = []
            for iClust in xrange(iNumClust):
                aClusters.append([])

            # calculate membership in clusters
            for iCard in xrange(iCards):
                oVec = aCards[iCard]
                iVmin = min(xrange(iNumClust),
                    key=lambda iV: fDist(oVec, aMeans[iV]))
                aClusters[iVmin].append(iCard)

            # recompute the centroids
            for iClust in xrange(iNumClust):
                if aClusters[iClust]:
                    aMeans[iClust] = (1.0 / len(aClusters[iClust])) * sum(
                        (aCards[x] for x in aClusters[iClust])
                    )

        return aMeans, aClusters

    def do_clustering(self):
        """Call the chosen clustering algorithm"""
        # gather cards
        aCards = list(self.model.get_card_iterator(None))

        # gather property functions
        dPropFuncs = {}
        for sGroup, dButtons in self._dPropButtons.iteritems():
            for sName, oBut in dButtons.iteritems():
                if oBut.get_active():
                    dPropFuncs[sGroup + ": " + sName] = \
                            self._dGroups[sGroup][sName]

        # sort column names
        aColNames = dPropFuncs.keys()
        aColNames.sort()

        # make tabulator and get table
        oTab = CardListTabulator(aColNames, dPropFuncs)
        aTable = oTab.tabulate(aCards)

        # set k-means parameters
        if self._oAutoNumClusters.get_active():
            iNumClusts = max(4, int(len(aTable) / 80.0) + 1)
        else:
            iNumClusts = max(2, int(self._oNumClustersSpin.get_value()))
        iIterations = max(2, int(self._oNumIterSpin.get_value()))
        for oBut in self._aDistanceMeasureGroup:
            if oBut.get_active():
                sName = oBut.get_label()
                fDist = Vector.METRICS[sName]
                break
        else:
            fDist = Vector.METRICS['Euclidean Distance']

        # aMeans -> list of vectors of cluster centroids
        # aClusters -> list of clusters, each cluster is a list of card indexes

        aMeans, aClusters = self.k_means(aTable, iNumClusts, iIterations,
                fDist)

        self._populate_results(aCards, aColNames, aMeans, aClusters)

    def _make_pcs_from_cluster(self, iClusterId):
        """Create a Card Set from the chosen cluster"""
        sDeckName = '_cluster_deck_%d' % (iClusterId,)

        # Check Deck Doesn't Exist
        if PhysicalCardSet.selectBy(name=sDeckName).count() != 0:
            do_complaint_error("Card Set %s already exists." %
                    sDeckName)
            return

        # Create Deck
        oDeck = PhysicalCardSet(name=sDeckName)

        # pylint: disable-msg=E1101
        # SQLObject confuses pylint
        for iIndex in self._aClusters[iClusterId]:
            oCard = IPhysicalCard(self._aCards[iIndex])
            oDeck.addPhysicalCard(oCard)

        self.open_cs(sDeckName, True)


class Vector(object):
    """Really simple class representing a row of card table data."""

    METRICS = {}

    def __init__(self, aData):
        self._aData = aData

    # We do want to access _aData on other Vectors.
    # pylint: disable-msg=W0212

    def euclidian_distance(self, oVec2):
        """Euclidean distance between two vectors."""
        assert len(self._aData) == len(oVec2._aData)
        fSum = sum(((x - y) ** 2 for (x, y) in zip(self._aData, oVec2._aData)))
        return math.sqrt(fSum)

    METRICS['Euclidean Distance'] = euclidian_distance

    def sutekh_distance(self, oVec2):
        """Like Euclidean distance, but -1 is distance 0.25 to anything
           (including 0 and -1) and 0 is distance 4.0 from anything other
           than 0.
           """
        assert len(self._aData) == len(oVec2._aData)
        fSum = sum((
            ((x == -1 or y == -1) and 0.25) or
            (((x == 0) ^ (y == 0) and 4.0)) or
            (x - y) ** 2 for (x, y) in zip(self._aData, oVec2._aData)
        ))
        return math.sqrt(fSum)

    METRICS['Sutekh Distance'] = sutekh_distance

    # Other possibile metrics:
    #  'City Block Distance',
    #  'Correlation',
    #  'Absolute Value of the Correlation',
    #  'Uncentered Correlation',
    #  'Absolute Value of the Uncentered Correlation',
    #  "Spearman's Rank Correlation"
    #  "Kendall's Tau"

    def __add__(self, oOther):
        """Sum this vector with another."""
        if oOther == 0:
            return self
        assert len(self._aData) == len(oOther._aData)
        return Vector([x + y for (x, y) in zip(self._aData, oOther._aData)])

    def __radd__(self, oOther):
        """Sum this vector with another."""
        if oOther == 0:
            return self
        assert len(self._aData) == len(oOther._aData)
        return Vector([x + y for (x, y) in zip(self._aData, oOther._aData)])

    def __mul__(self, fScale):
        """Multiply this vector by a scalar."""
        fScale = float(fScale)
        return Vector([x * fScale for x in self._aData])

    def __rmul__(self, fScale):
        """Multiply this vector by a scalar."""
        fScale = float(fScale)
        return Vector([x * fScale for x in self._aData])

    def __len__(self):
        """Length of this vector."""
        return len(self._aData)

    def __str__(self):
        """String representation of this vector."""
        return str(self._aData)

    def __iter__(self):
        """Iterator."""
        return iter(self._aData)

plugin = ClusterCardList



# test_WriteJOL.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Test Writing a card set to an JOL file"""

from sutekh.tests.TestCore import SutekhTest
from sutekh.tests.core.test_PhysicalCardSet import make_set_1
from sutekh.core.CardSetHolder import CardSetWrapper
from sutekh.io.WriteJOL import  WriteJOL
import unittest

EXPECTED_1 = """Abebe
Alan Sovereign (advanced)
Inez "Nurse216" Villagrande
The Siamese

4x.44 Magnum
2xAK-47
Aaron's Feeding Razor
2xAbbot
2xAbombwe
Scapelli, The Family "Mechanic"
The Path of Blood
"""


class JOLWriterTests(SutekhTest):
    """class for the JOL deck writer tests"""

    def test_deck_writer(self):
        """Test JOL deck writing"""
        oPhysCardSet1 = make_set_1()

        # Check output

        oWriter = WriteJOL()
        sData = self._round_trip_obj(oWriter, CardSetWrapper(oPhysCardSet1))

        self.assertEqual(sData, EXPECTED_1)


if __name__ == "__main__":
    unittest.main()



# CardSetImporter.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Simon Cross <hodgestar@gmail.com>,
# Copyright 2007 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Convert a ELDB or ARDB text or html file into an Card Set."""

import gtk
import urllib2
from sutekh.io.ELDBHTMLParser import ELDBHTMLParser
from sutekh.io.ARDBTextParser import ARDBTextParser
from sutekh.io.ARDBXMLDeckParser import ARDBXMLDeckParser
from sutekh.io.ARDBXMLInvParser import ARDBXMLInvParser
from sutekh.io.ELDBDeckFileParser import ELDBDeckFileParser
from sutekh.io.ELDBInventoryParser import ELDBInventoryParser
from sutekh.io.JOLDeckParser import JOLDeckParser
from sutekh.io.LackeyDeckParser import LackeyDeckParser
from sutekh.io.GuessFileParser import GuessFileParser
from sutekh.core.SutekhObjects import PhysicalCardSet
from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.SutekhDialog import SutekhDialog
from sutekh.gui.GuiCardSetFunctions import import_cs
from sutekh.gui.SutekhFileWidget import SutekhFileWidget


class ACSImporter(SutekhPlugin):
    """Convert a HTML or text deck into an ACS.

       Handles the most common formats, and allows the user to choose
       uri's, so decks published online can be easily imported.
       """
    dTableVersions = {PhysicalCardSet: (5, 6)}
    aModelsSupported = ("MainWindow",)

    # pylint: disable-msg=W0142
    # ** magic OK
    def __init__(self, *aArgs, **kwargs):
        super(ACSImporter, self).__init__(*aArgs, **kwargs)
        # Parser classes should be instantiable using Parser()
        # Parser objects should have a .parse(oFile, oCardSetHolder) method
        # (see the interface in IOBase)
        self._dParsers = {
            'ELDB HTML File': ELDBHTMLParser,
            'ARDB Text File': ARDBTextParser,
            'ELDB Deck (.eld)': ELDBDeckFileParser,
            'ELDB Inventory': ELDBInventoryParser,
            'ARDB Deck XML File': ARDBXMLDeckParser,
            'ARDB Inventory XML File': ARDBXMLInvParser,
            'JOL Deck File': JOLDeckParser,
            'Lackey CCG Deck File': LackeyDeckParser,
            'Guess file format': GuessFileParser,
        }
        self.oUri = None
        self.oDlg = None
        self._oFirstBut = None
        self.oFileChooser = None
        self._sNewName = ''

    def get_menu_item(self):
        """Register with the 'Import' Menu"""
        if not self.check_versions() or not self.check_model_type():
            return None
        oImport = gtk.MenuItem("Import Card Set in other formats")
        oImport.connect("activate", self.make_dialog)
        return ('Import Card Set', oImport)

    def make_dialog(self, _oWidget):
        """Create the dialog asking the user for the source to import."""
        self.oDlg = SutekhDialog("Choose Card Set File or URL", None,
                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_OK, gtk.RESPONSE_OK,
                 gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

        # pylint: disable-msg=E1101
        # vbox confuses pylint
        self.oDlg.vbox.pack_start(gtk.Label("URL:"), expand=False)

        self.oUri = gtk.Entry(150)
        self.oUri.connect("activate", self.handle_response, gtk.RESPONSE_OK)
        self.oDlg.vbox.pack_start(self.oUri, expand=False)

        self.oDlg.vbox.pack_start(gtk.Label("OR"), expand=False)

        self.oFileChooser = SutekhFileWidget(self.parent,
                gtk.FILE_CHOOSER_ACTION_OPEN)
        self.oFileChooser.add_filter_with_pattern('HTML files', ['*.html',
            '*.htm'])
        self.oFileChooser.add_filter_with_pattern('TXT files', ['*.txt'])
        self.oFileChooser.add_filter_with_pattern('ELD files', ['*.eld'])
        self.oFileChooser.add_filter_with_pattern('XML files', ['*.xml'])
        self.oFileChooser.add_filter_with_pattern('CSV files', ['*.csv'])
        self.oFileChooser.default_filter()
        self.oDlg.vbox.pack_start(self.oFileChooser)

        self._oFirstBut = gtk.RadioButton(None, 'Guess file format', False)
        self._oFirstBut.set_active(True)
        self.oDlg.vbox.pack_start(self._oFirstBut, expand=False)
        oTable = gtk.Table(len(self._dParsers) / 2, 2)
        self.oDlg.vbox.pack_start(oTable)
        oIter = self._dParsers.iterkeys()
        iXPos, iYPos = 0, 0
        for sName in oIter:
            if sName == 'Guess file format':
                continue
            oBut = gtk.RadioButton(self._oFirstBut, sName)
            oTable.attach(oBut, iXPos, iXPos + 1, iYPos, iYPos + 1)
            iXPos += 1
            if iXPos > 1:
                iXPos = 0
                iYPos += 1

        self.oDlg.connect("response", self.handle_response)
        self.oDlg.set_size_request(600, 500)
        self.oDlg.show_all()

        self.oDlg.run()

    def handle_response(self, _oWidget, oResponse):
        """Handle the user's clicking on OK or CANCEL in the dialog."""
        if oResponse == gtk.RESPONSE_OK:
            sUri = self.oUri.get_text().strip()
            sFile = self.oFileChooser.get_filename()

            for oBut in self._oFirstBut.get_group():
                sName = oBut.get_label()
                if oBut.get_active():
                    cParser = self._dParsers[sName]

            if sUri:
                self.make_cs_from_uri(sUri, cParser)
            elif sFile:
                self.make_cs_from_file(sFile, cParser)

        self.oDlg.destroy()

    def make_cs_from_uri(self, sUri, cParser):
        """From an URI, create an Card Set"""
        fIn = urllib2.urlopen(sUri)
        try:
            import_cs(fIn, cParser(), self.parent)
        finally:
            fIn.close()

    def make_cs_from_file(self, sFile, cParser):
        """From an file, create an Card Set"""
        fIn = file(sFile, "rb")
        try:
            import_cs(fIn, cParser(), self.parent)
        finally:
            fIn.close()


plugin = ACSImporter



# SutekhDialog.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2007 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details
# Provides base class for the dailogs used in Sutekh - ensures set_name
# is called consistently
# Also provides helpful wrapper functions around gtk.MessageDialog's

"""Dialog wrapper and functions for Sutekh"""

import gtk
import logging
import sys
import traceback
from gobject import markup_escape_text


class SutekhDialog(gtk.Dialog):
    # pylint: disable-msg=R0904
    # gtk widget, so has many public methods
    """wrapper class for gtk.Dialog"""
    def __init__(self, sTitle, oParent=None, iFlags=0, oButtons=None):
        super(SutekhDialog, self).__init__(sTitle, oParent, iFlags, oButtons)
        self.set_name("Sutekh.dialog")


def do_complaint(sMessage, oDialogType, oButtonType, bMarkup=False):
    """Wrapper function for gtk.MessageDialog.

       Create the dialog, run it, and return the result. If bMarkup is true,
       the string is interpreted as markup, other, just as plain text
       """
    if bMarkup:
        oComplaint = gtk.MessageDialog(None, 0, oDialogType,
                oButtonType, None)
        oComplaint.set_markup(sMessage)
    else:
        oComplaint = gtk.MessageDialog(None, 0, oDialogType,
                oButtonType, sMessage)
    oComplaint.set_name("Sutekh.dialog")
    iResponse = oComplaint.run()
    oComplaint.destroy()
    return iResponse


def do_complaint_buttons(sMessage, oType, aButtonInfo, bMarkup=False):
    """Wrapper function for gtk.MessageDialog, using add_button to create
       custom button layouts.

       Create the dialog, run it, and return the result. If bMarkup is true,
       the string is interpreted as markup, other, just as plain text.
       """
    if bMarkup:
        oComplaint = gtk.MessageDialog(None, 0, oType,
                gtk.BUTTONS_NONE, None)
        oComplaint.set_markup(sMessage)
    else:
        oComplaint = gtk.MessageDialog(None, 0, oType,
                gtk.BUTTONS_NONE, sMessage)
    for oItem, oResponse in zip(aButtonInfo[0::2], aButtonInfo[1::2]):
        oComplaint.add_button(oItem, oResponse)
    oComplaint.set_name("Sutekh.dialog")
    iResponse = oComplaint.run()
    oComplaint.destroy()
    return iResponse


def do_complaint_error(sMessage):
    """Error dialog with close button"""
    return do_complaint(sMessage, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, False)


def do_complaint_warning(sMessage):
    """Warning dialog with OK and CANCEL buttons"""
    return do_complaint(sMessage, gtk.MESSAGE_WARNING, gtk.BUTTONS_OK_CANCEL,
            False)


class DetailDialog(SutekhDialog):
    """Message dialog with a details expander"""
    # pylint: disable-msg=R0904
    # gtk widget, so has many public methods

    def __init__(self, sMessage, sDetails):
        super(DetailDialog, self).__init__('Sutekh has encounterd an error',
                oButtons=(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))
        oHBox = gtk.HBox(False, 2)
        oMessageBox = gtk.VBox(False, 2)
        oImage = gtk.Image()
        oImage.set_from_stock(gtk.STOCK_DIALOG_ERROR, gtk.ICON_SIZE_DIALOG)
        oImage.set_alignment(0, 0)
        oHBox.pack_start(oImage, expand=False)
        oInfo = gtk.Label()
        oInfo.set_markup('<b>%s</b>' % markup_escape_text(sMessage))
        oInfo.set_alignment(0, 0)
        oInfo.set_selectable(True)
        oMessageBox.pack_start(oInfo, expand=False)
        oExpander = gtk.Expander('Details')
        oFrame = gtk.Frame()
        oDetails = gtk.Label(sDetails)
        oDetails.set_selectable(True)
        oFrame.add(oDetails)
        oExpander.add(oFrame)
        oMessageBox.pack_start(oExpander)
        oHBox.pack_start(oMessageBox)
        self.vbox.pack_start(oHBox)
        oExpander.set_expanded(False)
        self.show_all()
        self.set_name("Sutekh.dialog")


def do_complaint_error_details(sMessage, sDetails):
    """Popup an details dialog for an error"""
    oComplaint = DetailDialog(sMessage, sDetails)
    iResponse = oComplaint.run()
    oComplaint.destroy()
    return iResponse


def do_exception_complaint(sMessage):
    """Handle an exception - log the details for verbose info, and popup
       a detailed dialog with the info."""
    oType, oValue, oTraceback = sys.exc_info()
    aTraceback = traceback.format_exception(oType, oValue, oTraceback,
            limit=30)
    logging.error("%s:\n%s", sMessage, "".join(aTraceback))
    do_complaint_error_details(sMessage, "".join(aTraceback))



# FilterModelPanes.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Handle the manipulation bits for the Filter Editor
# Copyright 2010 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Handle the panes for the filter editor"""

from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow
from sutekh.gui.ScrolledList import ScrolledList
from sutekh.gui.CustomDragIconView import CustomDragIconView
from sutekh.gui.CardSetsListView import CardSetsListView
from sutekh.core.FilterParser import get_filters_for_type
from sutekh.core.FilterBox import FilterBoxItem, FilterBoxModel, BOXTYPE, \
        BOXTYPE_ORDER
import gobject
import gtk
import pango

DRAG_TARGETS = [('STRING', 0, 0), ('text/plain', 0, 0)]
NO_DRAG_TARGET = [("No Drag", 0, 0)]
# Filters we pad values for to sort nicely
PAD_FILTERS = set(('Capacity', 'CardCount: Count'))
LIST_TYPES = set((FilterBoxItem.LIST, FilterBoxItem.LIST_FROM))

# Drag type constants
NEW_VALUE = 'NewValue: '
MOVE_VALUE = 'MoveValue: '
MOVE_FILTER = 'MoveFilter: '
NEW_FILTER = 'NewFilter: '

# Key constants
ENTER_KEYS = set([gtk.gdk.keyval_from_name('Return'),
    gtk.gdk.keyval_from_name('KP_Enter')])
LEFT_RIGHT = set([gtk.gdk.keyval_from_name('KP_Left'),
    gtk.gdk.keyval_from_name('KP_Right'),
    gtk.gdk.keyval_from_name('Left'),
    gtk.gdk.keyval_from_name('Right')])
LEFT = set([gtk.gdk.keyval_from_name('KP_Left'),
    gtk.gdk.keyval_from_name('Left')])
RIGHT = set([gtk.gdk.keyval_from_name('KP_Right'),
    gtk.gdk.keyval_from_name('Right')])


def unescape_markup(sMarkup):
    """Untiltiy function to strip markup from a string"""
    _oAttr, sStripped, _oAccel = pango.parse_markup(sMarkup)
    return sStripped


def add_accel_to_button(oButton, sAccelKey, oAccelGroup, sToolTip=None):
    """Creates a button using an gtk.AccelLabel to display the accelerator"""
    (iKeyVal, iMod) = gtk.accelerator_parse(sAccelKey)
    if iKeyVal != 0:
        oButton.add_accelerator('clicked', oAccelGroup, iKeyVal, iMod,
                gtk.ACCEL_VISIBLE)
    if sToolTip:
        oButton.set_tooltip_markup(sToolTip)


class FilterModelPanes(gtk.HBox):
    """Widget to hold the different panes of the Filter editor"""
    # pylint: disable-msg=R0904
    # R0904 - gtk.Widget, so many public methods

    def __init__(self, sFilterType, oDialog):
        super(FilterModelPanes, self).__init__()
        # Create the 3 panes
        self._oBoxModel = None
        self._sFilterType = sFilterType
        self._oSelectBar = FilterValuesBox(oDialog, sFilterType)
        self._oEditBox = FilterBoxModelEditBox(self._oSelectBar)

        self.pack_start(AutoScrolledWindow(self._oEditBox, True), expand=True)
        self.pack_start(self._oSelectBar, expand=True)

    def replace_ast(self, oAST):
        """Replace the AST in the tree model"""
        self._oBoxModel = FilterBoxModel(oAST, self._sFilterType)
        self._oEditBox.set_box_model(self._oBoxModel)

    def get_ast_with_values(self):
        """Get the current ast for the Editor, with values filled in"""
        if not self._oBoxModel:
            return None

        return self._oBoxModel.get_ast_with_values()

    def get_text(self):
        """Get the current ast"""
        if self._oBoxModel:
            return self._oBoxModel.get_text()
        return None


class FilterEditorToolbar(CustomDragIconView):
    """Toolbar listing the possible filter elements"""
    # pylint: disable-msg=R0904
    # R0904 - gtk.Widget, so many public methods

    def __init__(self, sFilterType):
        self._oListStore = gtk.ListStore(gobject.TYPE_STRING,
                gobject.TYPE_STRING)
        super(FilterEditorToolbar, self).__init__(self._oListStore)
        oTextCell = gtk.CellRendererText()
        oColumn = gtk.TreeViewColumn("Filter Element", oTextCell, text=0)
        oColumn.set_spacing(2)
        self.append_column(oColumn)
        self._sFilterType = sFilterType
        # Get supported filters
        aFilters = [('Filter Group ..', 'Filter Group')]
        for oFilterType in sorted(get_filters_for_type(self._sFilterType),
                key=lambda x: x.description):
            aFilters.append((oFilterType.description, oFilterType.keyword))
        self.drag_source_set(
                gtk.gdk.BUTTON1_MASK | gtk.gdk.BUTTON3_MASK,
                DRAG_TARGETS, gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE)
        # Create entries for each of the filters we support
        for tInfo in aFilters:
            self._oListStore.append(tInfo)

        self.connect('drag_data_get', self.drag_filter)
        self.get_selection().set_mode(gtk.SELECTION_SINGLE)

    def drag_filter(self, _oBtn, _oContext, oSelectionData, _oInfo, _oTime):
        """Create a drag info for this filter"""
        sSelect = self.get_paste_filter()
        if sSelect:
            oSelectionData.set(oSelectionData.target, 8, sSelect)

    def get_paste_filter(self):
        """Get the currently selected filter"""
        _oModel, oIter = self.get_selection().get_selected()
        if oIter:
            sSelect = '%s%s' % (NEW_FILTER,
                    self._oListStore.get_value(oIter, 1))
            return sSelect
        return None


class FilterValuesBox(gtk.VBox):
    """Holder for the value setting objects"""
    # pylint: disable-msg=R0904, R0902
    # R0904 - gtk.Widget, so many public methods
    # R0902: We need to keep a lot of state to handle all the cases

    def __init__(self, oDialog, sFilterType):
        super(FilterValuesBox, self).__init__()
        self._oEmptyWidget = gtk.VBox()
        self._oNoneWidget = gtk.VBox()
        self._oParent = oDialog
        self._sFilterType = sFilterType
        self._oEmptyWidget.pack_start(gtk.Label(
            'Select an active\nfilter element'), expand=False)
        self._oNoneWidget.pack_start(gtk.Label(
            'No values for this filter'), expand=False)
        # Handle removing none file elements
        self._set_drop_for_widget(self._oNoneWidget)
        self._oFilter = None
        self._oBoxModelEditor = None
        self._aLastSelection = []
        self._oLastFilter = None
        oCheckBox = gtk.HBox()

        self._oDisable = gtk.CheckButton('Disable (Ctrl-space)')
        self._oNegate = gtk.CheckButton('Negate (Alt-space)')
        self._oDelete = gtk.Button('Delete Filter or Value (del)')

        add_accel_to_button(self._oDisable, "<Ctl>space", oDialog.accel_group)
        add_accel_to_button(self._oNegate, "<Alt>space", oDialog.accel_group)
        add_accel_to_button(self._oDelete, "Delete", oDialog.accel_group)

        self._oDisable.set_sensitive(False)
        self._oDelete.set_sensitive(False)
        self._oNegate.set_sensitive(False)

        oCheckBox.pack_start(self._oDisable, expand=True)
        oCheckBox.pack_start(self._oNegate, expand=True)
        self.pack_start(oCheckBox, expand=False)
        self.pack_start(self._oDelete, expand=False)
        self._oDisable.connect('toggled', self.toggle_disabled)
        self._oDelete.connect('clicked', self.delete)
        self._oNegate.connect('toggled', self.toggle_negate)

        self._oWidget = self._oEmptyWidget
        self.pack_start(self._oWidget, expand=True)
        self.show_all()

    def _set_drop_for_widget(self, oWidget):
        """Set the correct drop dest behaviour for the given widget"""
        oViewWidget = self._get_view_widget(oWidget)
        oViewWidget.drag_dest_set(gtk.DEST_DEFAULT_ALL, DRAG_TARGETS,
                gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE)
        oViewWidget.connect('drag_data_received', self.drag_drop_handler)

    # pylint: disable-msg=R0201
    # Methods for consistency

    def _set_drag_for_widget(self, oWidget, fCallback, oFilter):
        """Set the correct drag source behaviour for the widget"""
        oViewWidget = self._get_view_widget(oWidget)
        oViewWidget.drag_source_set(
                gtk.gdk.BUTTON1_MASK | gtk.gdk.BUTTON3_MASK, DRAG_TARGETS,
                gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE)
        oViewWidget.connect('drag_data_get', fCallback, oFilter, oWidget)

    def _get_view_widget(self, oWidget):
        """Helper function to get view from widget if needed"""
        oSetWidget = oWidget
        if hasattr(oWidget, 'view'):
            oSetWidget = oWidget.view
        return oSetWidget

    # pylint: enable-msg=R0201

    def _make_filter_group_list(self, oFilter):
        """Create the toolbar for the filter group"""
        self._oWidget = gtk.VBox()
        oFilterTypes = ScrolledList('Filter Group Type')
        oFilterTypes.set_select_single()
        oFilterTypes.fill_list(BOXTYPE_ORDER)
        self.set_box_model_value(oFilter, oFilterTypes)
        oFilterTypes.get_selection_object().connect('changed',
                self.update_box_model, oFilter, oFilterTypes)
        oFilterTypes.set_size_request(100, 150)
        oFilterTypes.view.connect('key-press-event', self.key_press, oFilter,
                'Filter Type')
        self._oWidget.pack_start(oFilterTypes, expand=False)
        self._oWidget.pack_start(gtk.Label('Or Drag Filter Element'),
                expand=False)
        oSubFilterList = FilterEditorToolbar(self._sFilterType)
        oSubFilterList.connect('key-press-event', self.key_press, oFilter,
                'Filter')
        self._oWidget.pack_start(AutoScrolledWindow(oSubFilterList),
                    expand=True)

    def _make_list_from(self, oFilter):
        """Create the widget for the 'X form Y' filters"""
        self._oWidget = gtk.VBox()
        oCountList = ScrolledList('Select Counts', bSpecialSelect=True)
        oCountList.set_select_multiple()
        oCountList.fill_list(oFilter.aValues[0])
        oCountList.connect('key-press-event', self.key_press, oFilter, 'Count')
        oSetList = CardSetsListView(None, self._oParent, bSpecialSelect=True)
        oSetList.set_select_multiple()
        oSetList.connect('key-press-event', self.key_press, oFilter, 'Set')
        self._set_drag_for_widget(oSetList, self.update_count_set, oFilter)
        self._set_drop_for_widget(oSetList)
        self._set_drag_for_widget(oCountList, self.update_count_list, oFilter)
        self._set_drop_for_widget(oCountList)
        self._oWidget.pack_start(oCountList, expand=True)
        oLabel = gtk.Label()
        oLabel.set_markup('<b>From</b>')
        self._oWidget.pack_start(oLabel, expand=False)
        self._oWidget.pack_start(AutoScrolledWindow(oSetList, False),
                expand=True)

    def _make_card_set_list(self, oFilter):
        """Create a card set list widget"""
        oSetList = CardSetsListView(None, self._oParent, bSpecialSelect=True)
        oSetList.set_select_multiple()
        self._oWidget = AutoScrolledWindow(oSetList, False)
        oSetList.connect('key-press-event', self.key_press, oFilter, 'CardSet')
        self._set_drag_for_widget(oSetList, self.update_set_list, oFilter)
        self._set_drop_for_widget(oSetList)

    def _make_filter_values_list(self, oFilter):
        """Create a filter values list widget"""
        self._oWidget = ScrolledList('Select Filter Values',
                bSpecialSelect=True)
        self._oWidget.set_select_multiple()
        self._oWidget.fill_list(oFilter.aValues)
        self._oWidget.view.connect('key-press-event', self.key_press, oFilter,
                'Value')
        self._set_drag_for_widget(self._oWidget, self.update_filter_list,
                oFilter)
        self._set_drop_for_widget(self._oWidget)

    def _make_filter_entry(self, oFilter):
        """Create a text entry widget"""
        self._oWidget = gtk.VBox()
        self._set_drop_for_widget(self._oWidget)
        self._oWidget.pack_start(gtk.Label('Enter Text'), expand=False)
        oEntry = gtk.Entry()
        self._oWidget.pack_start(oEntry, expand=False)
        if oFilter.aCurValues:
            oEntry.set_text(oFilter.aCurValues[0])
        oEntry.connect('changed', self.update_edit_box, oFilter)
        oEntry.connect('key-press-event', self.key_press, oFilter, 'Entry')

    def set_widget(self, oFilter, oBoxModelEditor):
        """Replace the current widget with the correct widget for the
           new filter"""
        self._oBoxModelEditor = oBoxModelEditor
        self._aLastSelection = []
        if self._oWidget:
            if self._oLastFilter == oFilter:
                return  # We're still the same filter, so do nothing
            self.remove(self._oWidget)
            self._oWidget = None
        if isinstance(oFilter, FilterBoxModel):
            # Select between box options
            self._oDisable.set_active(oFilter.bDisabled)
            self._oNegate.set_inconsistent(True)
            self._oNegate.set_sensitive(False)
            self._make_filter_group_list(oFilter)
        elif isinstance(oFilter, FilterBoxItem):
            self._oDisable.set_active(oFilter.bDisabled)
            self._oNegate.set_inconsistent(False)
            self._oNegate.set_active(oFilter.bNegated)
            self._oNegate.set_sensitive(True)
            if oFilter.iValueType == FilterBoxItem.LIST:
                # Select appropriate list widget for this filter
                if oFilter.sFilterName == "Card_Sets" or \
                        oFilter.sFilterName == "ParentCardSet":
                    # Special case to use card set list widget
                    self._make_card_set_list(oFilter)
                else:
                    # Ordinary list
                    self._make_filter_values_list(oFilter)
            elif oFilter.iValueType == FilterBoxItem.ENTRY:
                self._make_filter_entry(oFilter)
            elif oFilter.iValueType == FilterBoxItem.LIST_FROM:
                self._make_list_from(oFilter)
            elif oFilter.iValueType == FilterBoxItem.NONE:
                # None filter, so no selection widget, but we have buttons
                self._oWidget = self._oNoneWidget
                self._oWidget.connect('key-press-event', self.key_press, None,
                        'None')
        else:
            # No selected widget, so clear everything
            self._oWidget = self._oEmptyWidget
            self.disable_all_buttons()
        self.pack_start(self._oWidget, expand=True)
        self._oLastFilter = oFilter
        self.show_all()

    # pylint: disable-msg=R0201
    # Methods for consistency
    def set_box_model_value(self, oBoxModel, oWidget):
        """Set the correct selection for this box model"""
        for sDesc, tInfo in BOXTYPE.iteritems():
            sBoxType, bNegate = tInfo
            if oBoxModel.sBoxType == sBoxType and oBoxModel.bNegate == bNegate:
                oWidget.set_selected_entry(sDesc)

    # pylint: enable-msg=R0201

    def update_box_model(self, _oSelection, oBoxModel, oList):
        """Update the box model to the current selection"""
        aSelected = oList.get_selection()
        if not aSelected:
            return  # We don't do anything special if nothing's selected
        for sDesc, tInfo in BOXTYPE.iteritems():
            if sDesc == aSelected[0]:
                oBoxModel.sBoxType, oBoxModel.bNegate = tInfo
        self._oBoxModelEditor.update_box_text(oBoxModel)

    # pylint: disable-msg=R0913
    # function signature requires all these arguments

    def drag_drop_handler(self, _oWindow, oDragContext, _iXPos, _iYPos,
            oSelectionData, _oInfo, oTime):
        """Handle drops from the filter toolbar"""
        bDragRes = True
        if not oSelectionData and oSelectionData.format != 8:
            bDragRes = False
        else:
            sData = oSelectionData.data
            if not sData:
                bDragRes = False
            elif sData.startswith(MOVE_VALUE):
                # Removing a value from the list
                sIter = sData.split(':', 1)[1].strip()
                oIter = self._oBoxModelEditor.get_iter_from_string(sIter)
                self._oBoxModelEditor.remove_value_at_iter(oIter)
            elif sData.startswith(MOVE_FILTER):
                # Removing a filter
                sIter = sData.split(':', 1)[1].strip()
                oIter = self._oBoxModelEditor.get_iter_from_string(sIter)
                self._oBoxModelEditor.remove_filter_at_iter(oIter)
            else:
                # Not relevant
                bDragRes = False
        oDragContext.finish(bDragRes, False, oTime)

    # pylint: enable-msg=R0913

    def get_selected_values(self):
        """Get the most recent selection - used by the drop values code"""
        return self._aLastSelection

    def _format_selection(self, sSelect, aSelection):
        """Set the current selection, properly formatted for sorting"""
        sName = sSelect.replace(NEW_VALUE, '')
        if sName in PAD_FILTERS:
            # Is this the best approach?
            self._aLastSelection = ['%2s' % x for x in aSelection]
        else:
            self._aLastSelection = aSelection

    def update_filter_list(self, _oBtn, _oContext, oSelectionData,
            _oInfo, _oTime, oFilter, _oWidget):
        """Update the box model with the new values"""
        sSelect = '%s%s' % (NEW_VALUE, oFilter.sFilterName)
        self._format_selection(sSelect, self._oWidget.get_selection())
        oSelectionData.set(oSelectionData.target, 8, sSelect)

    def update_count_list(self, _oBtn, _oContext, oSelectionData,
            _oInfo, _oTime, oFilter, oCountList):
        """Update the box model with the new values"""
        sSelect = '%s%s: Count' % (NEW_VALUE, oFilter.sFilterName)
        self._format_selection(sSelect, oCountList.get_selection())
        oSelectionData.set(oSelectionData.target, 8, sSelect)

    def update_set_list(self, _oBtn, _oContext, oSelectionData,
            _oInfo, _oTime, oFilter, oSetList):
        """Update the box model to the current selection"""
        sSelect = '%s%s' % (NEW_VALUE, oFilter.sFilterName)
        self._format_selection(sSelect, oSetList.get_all_selected_sets())
        oSelectionData.set(oSelectionData.target, 8, sSelect)

    def update_count_set(self, _oBtn, _oContext, oSelectionData,
            _oInfo, _oTime, oFilter, oSetList):
        """Update the box model to the current selection"""
        sSelect = '%s%s: Set' % (NEW_VALUE, oFilter.sFilterName)
        self._format_selection(sSelect, oSetList.get_all_selected_sets())
        oSelectionData.set(oSelectionData.target, 8, sSelect)

    def update_edit_box(self, oEntry, oFilter):
        """Update the box model with the current text"""
        oFilter.aCurValues = [oEntry.get_text()]
        self._oBoxModelEditor.update_entry(oFilter)

    def toggle_negate(self, oButton):
        """Toggle the disabled flag for a section of the filter"""
        if self._oBoxModelEditor:
            self._oBoxModelEditor.set_negate(oButton.get_active())

    def switch_focus(self, oEvent, sSource):
        """Handle switching focus between the different filter bits"""
        if sSource in set(['Value', 'Entry', 'None', 'CardSet']):
            # Only one entry in the filter pane
            self._oBoxModelEditor.grab_focus()
        elif sSource == 'Editor':
            if isinstance(self._oLastFilter, FilterBoxItem):
                if self._oLastFilter.iValueType == FilterBoxItem.LIST:
                    # Only one entry in the filter pane
                    self._oWidget.view.grab_focus()
                elif self._oLastFilter.iValueType == FilterBoxItem.ENTRY:
                    # entry is the second child of the VBox
                    self._oWidget.get_children()[1].grab_focus()
            if isinstance(self._oLastFilter, FilterBoxModel) or \
                    (isinstance(self._oLastFilter, FilterBoxItem) and
                            self._oLastFilter.iValueType ==
                            FilterBoxItem.LIST_FROM):
                if oEvent.keyval in LEFT:
                    # bottom child needs the focus
                    oAutoScrolled = self._oWidget.get_children()[2]
                    oAutoScrolled.get_child().grab_focus()
                else:
                    # top child needs the focus
                    self._oWidget.get_children()[0].view.grab_focus()
        elif sSource in set(['Count', 'Set', 'Filter Type', 'Filter']):
            if oEvent.keyval in RIGHT and \
                    sSource in set(['Count', 'Filter Type']):
                # Top child selected, so pass to bottom
                oAutoScrolled = self._oWidget.get_children()[2]
                oAutoScrolled.get_child().grab_focus()
            elif oEvent.keyval in LEFT and \
                    sSource in set(['Filter', 'Set']):
                # Bottom child selected, so pass to top
                self._oWidget.get_children()[0].view.grab_focus()
            else:
                # Passing focus to Model Editor
                self._oBoxModelEditor.grab_focus()

    def key_press(self, oWidget, oEvent, oFilter, sSource):
        """Handle key press events to do allow keyboard pasting"""
        # Alt-direction moves focus around
        if oEvent.keyval in LEFT_RIGHT and \
                (oEvent.get_state() & gtk.gdk.MOD1_MASK):
            self.switch_focus(oEvent, sSource)
        # We flag on ctrl-enter
        if oEvent.keyval not in ENTER_KEYS or not \
                (oEvent.get_state() & gtk.gdk.CONTROL_MASK):
            return
        if sSource == 'Filter':
            # Paste current filter into the current filter box
            sSelect = oWidget.get_paste_filter()
            if sSelect:
                self._oBoxModelEditor.paste_filter(oFilter, sSelect, -1)
        elif sSource == 'Count':
            sSelect = '%s%s: Count' % (NEW_VALUE, oFilter.sFilterName)
            self._format_selection(sSelect, oWidget.get_selection())
            self._oBoxModelEditor.paste_value(sSelect)
        elif sSource == 'Set':
            sSelect = '%s%s: Set' % (NEW_VALUE, oFilter.sFilterName)
            self._format_selection(sSelect, oWidget.get_all_selected_sets())
            self._oBoxModelEditor.paste_value(sSelect)
        elif sSource == 'CardSet':
            sSelect = '%s%s' % (NEW_VALUE, oFilter.sFilterName)
            self._format_selection(sSelect, oWidget.get_all_selected_sets())
            self._oBoxModelEditor.paste_value(sSelect)
        elif sSource == 'Value':
            sSelect = '%s%s' % (NEW_VALUE, oFilter.sFilterName)
            self._format_selection(sSelect, self._oWidget.get_selection())
            self._oBoxModelEditor.paste_value(sSelect)

    def toggle_disabled(self, oButton):
        """Toggle the disabled flag for a section of the filter"""
        if self._oBoxModelEditor:
            self._oBoxModelEditor.set_disabled(oButton.get_active())

    def delete(self, _oButton):
        """Delete a filter element"""
        if self._oBoxModelEditor:
            self._oBoxModelEditor.delete(None)

    def disable_all_buttons(self):
        """Disable all the buttons"""
        self._oDisable.set_sensitive(False)
        self._oDelete.set_sensitive(False)
        self._oNegate.set_sensitive(False)

    def enable_delete(self):
        """Enable the delete button"""
        self._oDelete.set_sensitive(True)

    def enable_disable(self, oFilterObj):
        """Enable the delete button"""
        self._oDisable.set_sensitive(True)
        self._oDisable.set_active(oFilterObj.bDisabled)

    def get_current_pos_and_sel(self):
        """Get the current scroll position and selection path in the
           sub-filter list to restore later"""

        def _get_values(oAdj):
            """Get important values out of a gtk.Adjustment"""
            return oAdj.value, oAdj.page_size

        # We must always be calling this with a filter box model shown
        assert len(self._oWidget.get_children()) == 3
        oScrolledWindow = self._oWidget.get_children()[2]
        oSubFilterWidget = oScrolledWindow.get_children()[0]
        _oModel, aPaths = oSubFilterWidget.get_selection().get_selected_rows()
        # We use the values, rather than the actual adjustments, to avoid a
        # race condition where gtk deletes the actual adjustment out from
        # under us
        tHorizVals = _get_values(oScrolledWindow.get_hadjustment())
        tVertVals = _get_values(oScrolledWindow.get_vadjustment())
        return tHorizVals, tVertVals, aPaths

    def restore_pos_and_selection(self, tScrollAdj):
        """Restore the selection and scrollbar position in the sub-filter
           list"""
        def _set_values(oAdj, tInfo):
            """Set the values on an adjustment"""
            oAdj.value = tInfo[0]
            oAdj.page_size = tInfo[1]
            # Send required signals
            oAdj.changed()
            oAdj.value_changed()

        # FIXME: fix selection behaviour so we select the correct
        # filter box model in this case, rather than the newly added
        # filter.
        if len(self._oWidget.get_children()) != 3:
            # Selected widget is not a box model (can happen if we insert
            # something before the box model, so no sensible default selection
            return
        oScrolledWindow = self._oWidget.get_children()[2]
        oSubFilterWidget = oScrolledWindow.get_children()[0]
        oSelection = oSubFilterWidget.get_selection()
        # selection will be a list with 1 path
        oSelection.select_path(tScrollAdj[2][0])
        _set_values(oScrolledWindow.get_hadjustment(), tScrollAdj[0])
        _set_values(oScrolledWindow.get_vadjustment(), tScrollAdj[1])


class BoxModelPopupMenu(gtk.Menu):
    """Popup context menu for disable/ negate & delete"""
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods

    def __init__(self, oBoxModelEditor):
        super(BoxModelPopupMenu, self).__init__()
        self._oDis = gtk.MenuItem("Disable / Enable Filter")
        self._oNeg = gtk.MenuItem("Negate Filter Element")
        self._oDel = gtk.MenuItem("Delete filter or value")

        self._oDis.connect("activate", oBoxModelEditor.toggle_disabled)
        self._oNeg.connect("activate", oBoxModelEditor.toggle_negate)
        self._oDel.connect("activate", oBoxModelEditor.delete)
        self.append(self._oDis)
        self.append(self._oNeg)
        self.append(gtk.SeparatorMenuItem())
        self.append(self._oDel)


class FilterBoxModelStore(gtk.TreeStore):
    """TreeStore for the FilterBoxModelEditor"""
    # pylint: disable-msg=R0904
    # R0904 - gtk.Widget, so many public methods

    BLACK = gtk.gdk.color_parse('black')
    GREY = gtk.gdk.color_parse('grey')
    NO_VALUE = '<i>No Values Set</i>'
    NONE_VALUE = '<b>No Values for this filter</b>'

    def __init__(self):
        super(FilterBoxModelStore, self).__init__(gobject.TYPE_STRING,
                gobject.TYPE_PYOBJECT, gtk.gdk.Color)

    def _fill_values(self, oFilterIter, oModel, oColour):
        """Fill in the values for this filter"""
        if oModel.iValueType == oModel.LIST_FROM:
            # Load LIST_FROM
            aValues, aFrom = oModel.aCurValues
            if aValues:
                for sValue in aValues:
                    self.append(oFilterIter,
                            (gobject.markup_escape_text(sValue), None,
                                oColour))
            else:
                self.append(oFilterIter, (self.NO_VALUE, None, oColour))
            self.append(oFilterIter, ('<b>From</b>', None, oColour))
            if aFrom:
                for sValue in aFrom:
                    self.append(oFilterIter,
                            (gobject.markup_escape_text(sValue), None,
                                oColour))
            else:
                self.append(oFilterIter, (self.NO_VALUE, None, oColour))
        elif oModel.aCurValues and oModel.aValues:
            for sValue in oModel.aCurValues:
                self.append(oFilterIter,
                        (gobject.markup_escape_text(sValue), None, oColour))
        elif oModel.aCurValues:
            self.append(oFilterIter,
                    (gobject.markup_escape_text(oModel.aCurValues[0]), None,
                        oColour))
        elif oModel.iValueType == oModel.NONE:
            self.append(oFilterIter, (self.NONE_VALUE, None, oColour))
        else:
            self.append(oFilterIter, (self.NO_VALUE, None, oColour))

    def fix_filter_name(self, oIter, oModel):
        """Fix the name for the given filter to reflect it negated state"""
        if oModel.bNegated:
            sText = 'NOT %s' % oModel.sFilterDesc
        else:
            sText = oModel.sFilterDesc
        self.set(oIter, 0, sText)

    def _do_add_iter(self, oIter, oModel, bDisabled):
        """Recursively add elements of the box model"""
        oThisIter = self.append(oIter)
        oColour = self.BLACK
        if hasattr(oModel, 'sFilterDesc'):
            if oModel.bDisabled or bDisabled:
                oColour = self.GREY
            self.set(oThisIter, 1, oModel, 2, oColour)
            self.fix_filter_name(oThisIter, oModel)
            self._fill_values(oThisIter, oModel, oColour)
        elif hasattr(oModel, 'sBoxType'):
            # Box Model, so lookup correct string
            if oModel.bDisabled or bDisabled:
                oColour = self.GREY
                bDisabled = True
            for sDesc, tInfo in BOXTYPE.iteritems():
                sBoxType, bNegate = tInfo
                if oModel.sBoxType == sBoxType \
                        and oModel.bNegate == bNegate:
                    self.set(oThisIter, 0, sDesc, 1, oModel, 2, oColour)
            # iterate over children
            for oSubModel in oModel:
                self._do_add_iter(oThisIter, oSubModel, bDisabled)

    def load(self, oBoxModel):
        """Load the box model into the store"""
        self.clear()
        if oBoxModel is None:
            return
        # Walk the box model, creating items as we need them
        self._do_add_iter(None, oBoxModel, False)
        return self.get_path(self.get_iter_root())

    def _clear_iter(self, oSelectIter, oPath):
        """Clear the current iter, saving curently selected element"""
        oChild = self.iter_children(oSelectIter)
        oCurIter = self.get_iter(oPath)
        iPos = 0
        iIndex = -1
        while oChild:
            oNext = self.iter_next(oChild)
            if not self.get_path(oChild) == self.get_path(oCurIter):
                self.remove(oChild)
            else:
                iIndex = iPos
            oChild = oNext
            iPos += 1
        if iIndex < 0:
            # We deleted everything, so add 1 child
            oCurIter = self.append(oSelectIter)
            iIndex = 0
        return iIndex, oCurIter

    def _add_list_iter(self, iPos, iIndex, oSelectIter, oCurIter):
        """Set the iter correctly for the list update"""
        if iPos == iIndex:
            oChild = oCurIter
        elif iPos < iIndex:
            oChild = self.insert_before(oSelectIter, oCurIter)
        else:
            oChild = self.append(oSelectIter)
        return oChild

    def update_iter_with_values(self, aValues, oSelectIter, oPath):
        """Update the given iter with the changed values"""
        iIndex, oCurIter = self._clear_iter(oSelectIter, oPath)
        if aValues:
            iIndex = min(len(aValues) - 1, iIndex)
            iPos = 0
            for sValue in aValues:
                oChild = self._add_list_iter(iPos, iIndex, oSelectIter,
                        oCurIter)
                if sValue is not None:
                    self.set(oChild, 0, gobject.markup_escape_text(sValue),
                            1, None, 2, self.BLACK)
                else:
                    self.set(oChild, 0, self.NO_VALUE, 1, None, 2, self.BLACK)
                iPos += 1
        else:
            self.set(oCurIter, 0, self.NO_VALUE, 1, None, 2, self.BLACK)
        return self.get_path(oCurIter)

    def update_iter_with_from_list(self, aCounts, aNames, oSelectIter, oPath):
        """Update the given iter with the changed values"""
        iIndex, oCurIter = self._clear_iter(oSelectIter, oPath)
        iIndex = min(len(aCounts) + len(aNames), iIndex)
        iPos = 0
        for sValue in aCounts:
            oChild = self._add_list_iter(iPos, iIndex, oSelectIter, oCurIter)
            if sValue is not None:
                self.set(oChild, 0, gobject.markup_escape_text(sValue),
                        1, None, 2, self.BLACK)
            else:
                self.set(oChild, 0, self.NO_VALUE, 1, None, 2, self.BLACK)
            iPos += 1
        oChild = self._add_list_iter(iPos, iIndex, oSelectIter, oCurIter)
        iPos += 1
        self.set(oChild, 0, '<b>From</b>', 1, None, 2, self.BLACK)
        for sValue in aNames:
            oChild = self._add_list_iter(iPos, iIndex, oSelectIter, oCurIter)
            if sValue is not None:
                self.set(oChild, 0, gobject.markup_escape_text(sValue),
                        1, None, 2, self.BLACK)
            else:
                self.set(oChild, 0, self.NO_VALUE, 1, None, 2, self.BLACK)
            iPos += 1
        return self.get_path(oCurIter)

    def update_entry(self, oFilterItem, oSelectIter):
        """Update the store with the current text"""
        oChild = self.iter_children(oSelectIter)
        if not oChild:
            # No children, so oSelectIter is the one we want
            oChild = oSelectIter
        sText = oFilterItem.aCurValues[0]
        if not sText:
            sText = self.NO_VALUE
        else:
            sText = gobject.markup_escape_text(sText)
        self.set(oChild, 0, sText, 1, None)
        return self.get_path(oChild)

    def get_drop_iter(self, tRowInfo):
        """Get the correct iter from the drop info"""
        iIndex = 0
        if not tRowInfo:
            oIter = self.get_iter_root()
            iIndex = -1
        else:
            oPath, iDropPos = tRowInfo
            oIter = self.get_iter(oPath)
            if self.iter_depth(oIter) > 0 and \
                    iDropPos == gtk.TREE_VIEW_DROP_BEFORE:
                # Find the iter immediately before this one, since
                # that's our actual target
                oParIter = self.iter_parent(oIter)
                oPrevIter = oParIter
                oNextIter = self.iter_children(oParIter)
                while self.get_path(oNextIter) != oPath:
                    oPrevIter = oNextIter
                    oNextIter = self.iter_next(oPrevIter)
                oIter = oPrevIter
        # oIter now points to the right point to insert
        return iIndex, oIter

    def get_drop_filter(self, tRowInfo):
        """Get the filter associated with the given drop point"""
        _iIndex, oIter = self.get_drop_iter(tRowInfo)
        oFilterObj = self.get_value(oIter, 1)
        while oFilterObj is None:
            oIter = self.iter_parent(oIter)
            oFilterObj = self.get_value(oIter, 1)
        return oFilterObj, oIter

    def get_insert_box(self, tRowInfo):
        """Get the filter box model to insert into."""
        iIndex, oIter = self.get_drop_iter(tRowInfo)
        oFilterObj = self.get_value(oIter, 1)
        if not hasattr(oFilterObj, 'sBoxType'):
            # Need to insert into parent box model
            oTempIter = oIter
            oInsertObj = oFilterObj
            while not hasattr(oInsertObj, 'sBoxType'):
                oTempIter = self.iter_parent(oTempIter)
                oInsertObj = self.get_value(oTempIter, 1)
                if oFilterObj is None:
                    # Bounce this up a level as well
                    oFilterObj = oInsertObj
            # We insert after this filter
            iIndex = oInsertObj.index(oFilterObj) + 1
        else:
            oInsertObj = oFilterObj
        return oInsertObj, iIndex


class FilterBoxModelEditView(CustomDragIconView):
    """TreeView for the FilterBoxModelEditor"""
    # pylint: disable-msg=R0904
    # R0904 - gtk.Widget, so many public methods

    def __init__(self, oStore, oValuesWidget, oBoxModel):
        super(FilterBoxModelEditView, self).__init__(oStore)
        self._oStore = oStore
        self._oBoxModel = oBoxModel
        oTextCell = gtk.CellRendererText()
        oColumn = gtk.TreeViewColumn("Filter", oTextCell, markup=0,
                foreground_gdk=2)
        self.append_column(oColumn)

        self.drag_dest_set(gtk.DEST_DEFAULT_ALL, DRAG_TARGETS,
                gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE)
        self.drag_source_set(
                gtk.gdk.BUTTON1_MASK | gtk.gdk.BUTTON3_MASK,
                DRAG_TARGETS, gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE)

        oSelection = self.get_selection()
        oSelection.set_mode(gtk.SELECTION_SINGLE)
        self._oValuesWidget = oValuesWidget
        self.oCurSelectIter = None
        oSelection.connect('changed', self.update_values_widget)

        self.connect('drag_data_received', self.drag_drop_handler)
        self.connect('drag_data_get', self.drag_element)
        self.connect('button_press_event', self.press_button)
        self.connect('key_press_event', self.key_press)

    def key_press(self, _oWidget, oEvent):
        """Handle key-press events for the filter box model"""
        if oEvent.keyval in LEFT_RIGHT and \
                (oEvent.get_state() & gtk.gdk.MOD1_MASK):
            # Pass focus to the value widget
            self._oValuesWidget.switch_focus(oEvent, 'Editor')

    def update_values_widget(self, _oTreeSelection):
        """Update the values widget to the new selection"""
        # Get the current selected row
        _oModel, oIter = self.get_selection().get_selected()
        oFilterObj = None
        if oIter:
            oFilterObj = self._oStore.get_value(oIter, 1)
            while oFilterObj is None:
                oIter = self._oStore.iter_parent(oIter)
                if oIter:
                    oFilterObj = self._oStore.get_value(oIter, 1)
        self.oCurSelectIter = oIter
        if oFilterObj is None:
            self._oValuesWidget.set_widget(None, self)
            return
        elif oFilterObj.bDisabled:
            self._oValuesWidget.set_widget(None, self)
        else:
            self._oValuesWidget.set_widget(oFilterObj, self)
        if self._oStore.iter_depth(self.oCurSelectIter) > 0:
            self._oValuesWidget.enable_disable(oFilterObj)
            self._oValuesWidget.enable_delete()
        else:
            self._oValuesWidget.disable_all_buttons()

    def select_path(self, oPath):
        """Helper function to manage setting the selected path"""
        self.set_cursor(oPath)
        self.grab_focus()
        self.scroll_to_cell(oPath, None, True, 0.5, 0.0)

    def set_box_model(self, oBoxModel):
        """Set the box model to the correct value"""
        self._oBoxModel = oBoxModel
        self.load()

    def load(self, oPath=None):
        """Load the boxmodel into the TreeView"""
        oRootPath = self._oStore.load(self._oBoxModel)
        self.expand_all()
        if oPath:
            self.select_path(oPath)
        else:
            # Select the root of the model by default.
            self.select_path(oRootPath)

    def get_iter_from_string(self, sIter):
        """Lookup the iter given the string. Used used dragging stuff out
           of the pane."""
        return self._oStore.get_iter_from_string(sIter)

    def remove_value_at_iter(self, oIter):
        """Remove the filter value at the given point in the tree"""
        oFilter = self._oStore.get_value(self._oStore.iter_parent(oIter), 1)
        sValue = unescape_markup(self._oStore.get_value(oIter, 0))
        if oFilter.iValueType == FilterBoxItem.LIST \
                and sValue in oFilter.aCurValues:
            oFilter.aCurValues.remove(sValue)
            self.update_list(self.oCurSelectIter, oFilter)
        elif oFilter.iValueType == FilterBoxItem.LIST_FROM:
            if oFilter.aCurValues[0] and \
                    sValue in oFilter.aCurValues[0]:
                oFilter.aCurValues[0].remove(sValue)
            elif oFilter.aCurValues[1] and \
                    sValue in oFilter.aCurValues[1]:
                oFilter.aCurValues[1].remove(sValue)
            self.update_count_list(self.oCurSelectIter, oFilter)

    def remove_filter_at_iter(self, oIter):
        """Remove the filter element at the given point in the tree"""
        oMoveObj = self._oStore.get_value(oIter, 1)
        oParent = self._oStore.get_value(self._oStore.iter_parent(oIter), 1)
        oParent.remove(oMoveObj)
        self.load()  # May break stuff

    def update_list(self, oIter, oFilterItem):
        """Update the list to show the current values"""
        self._update_iter_with_list(oIter, oFilterItem.aCurValues)

    def update_count_list(self, oIter, oFilterItem):
        """Update the list to show the current values"""
        aCounts = []
        aNames = []
        if oFilterItem.aCurValues[0]:
            aCounts.extend(oFilterItem.aCurValues[0])
        else:
            aCounts.append(None)
        if oFilterItem.aCurValues[1]:
            aNames.extend(oFilterItem.aCurValues[1])
        else:
            aNames.append(None)
        self._update_iter_with_from_list(oIter, aCounts, aNames)

    def update_box_text(self, oBoxModel):
        """Update the listing for the given box model"""
        for sDesc, tInfo in BOXTYPE.iteritems():
            sBoxType, bNegate = tInfo
            if oBoxModel.sBoxType == sBoxType and oBoxModel.bNegate == bNegate:
                self._oStore.set(self.oCurSelectIter, 0, sDesc)

    def update_entry(self, oFilterItem):
        """Update the filter with the current text"""
        oChildPath = self._oStore.update_entry(oFilterItem,
                self.oCurSelectIter)
        self.expand_to_path(oChildPath)
        self.expand_row(oChildPath, True)

    def _update_iter_with_list(self, oIter, aValues):
        """Fill in the list values"""
        oPath, _oCol = self.get_cursor()
        oCurPath = self._oStore.update_iter_with_values(aValues,
                oIter, oPath)
        self.expand_to_path(oCurPath)
        self.expand_row(oCurPath, True)

    def _update_iter_with_from_list(self, oIter, aCounts, aNames):
        """Fill in with a from list"""
        oPath, _oCol = self.get_cursor()
        oCurPath = self._oStore.update_iter_with_from_list(aCounts,
                aNames, oIter, oPath)
        self.expand_to_path(oCurPath)
        self.expand_row(oCurPath, True)

    def _check_for_obj(self, _oModel, oPath, oIter, oFilterObj):
        """Helper function for selecting an object in the tree.
          Meant to be called via the foreach method."""
        oCurObj = self._oStore.get_value(oIter, 1)
        if oCurObj is oFilterObj:
            self.select_path(oPath)

    def _check_for_value(self, _oModel, oPath, oIter, tValueInfo):
        """Helper function for selecting a value in a filter in the tree.
          Meant to be called via the foreach method."""
        oFilterObj, sValue = tValueInfo
        sCurValue = unescape_markup(self._oStore.get_value(oIter, 0))
        if sCurValue == sValue:
            # Check parent
            oParIter = self._oStore.iter_parent(oIter)
            oCurFilter = self._oStore.get_value(oParIter, 1)
            if oCurFilter is oFilterObj:
                self.select_path(oPath)

    # pylint: disable-msg=R0913
    # function signature requires all these arguments

    def drag_drop_handler(self, _oWindow, oDragContext, iXPos, iYPos,
            oSelectionData, _oInfo, oTime):
        """Handle drops from the filter toolbar"""
        bDragRes = True
        if not oSelectionData and oSelectionData.format != 8:
            bDragRes = False
        else:
            sData = oSelectionData.data
            tRowInfo = self.get_dest_row_at_pos(iXPos, iYPos)
            if not sData:
                bDragRes = False
            elif sData.startswith(NEW_FILTER) or \
                    sData.startswith(MOVE_FILTER):
                if not self._do_drop_filter(sData, tRowInfo):
                    bDragRes = False
            elif sData.startswith(MOVE_VALUE):
                if not self._do_move_value(sData, tRowInfo):
                    bDragRes = False
            elif sData.startswith(NEW_VALUE):
                if not self._do_drop_value(sData, tRowInfo):
                    bDragRes = False
            else:
                bDragRes = False
        oDragContext.finish(bDragRes, False, oTime)

    def paste_value(self, sData):
        """Paste data from the widgets"""
        if sData.startswith(NEW_VALUE):
            oFilterObj, _oPath = self._get_cur_filter()
            self._update_values(oFilterObj, self.oCurSelectIter, sData)

    def _check_drag_value(self, oIter, oFilter, tCurTarget, oAction,
            oDragContext):
        """Check if the current target is an acceptable drop location
           for a filter value"""
        if not tCurTarget:
            oDragContext.drag_status(0)
            return False
        if oFilter is None:
            oFilter = self._oStore.get_value(
                    self._oStore.iter_parent(oIter), 1)
        # Check that destination is valid for this value
        oDropObj, _oIter = self._oStore.get_drop_filter(tCurTarget)
        if hasattr(oDropObj, 'sFilterName') and \
               oDropObj.sFilterName == oFilter.sFilterName:
            oDragContext.drag_status(oAction)
            return True
        else:
            oDragContext.drag_status(0)
            return False

    def _fix_highlight(self, tCurTarget):
        """Fix the highlighted row to match the actual drop target for
           filters"""
        _iIndex, oIter = self._oStore.get_drop_iter(tCurTarget)
        oDropObj = self._oStore.get_value(oIter, 1)
        if oDropObj is None:
            # Target is a filter value, so set highlight to the
            # parent filter
            oDropPath = self._oStore.get_path(self._oStore.iter_parent(oIter))
            self.set_drag_dest_row(oDropPath, gtk.TREE_VIEW_DROP_AFTER)

    def drag_motion(self, _oWidget, oDragContext, iXPos, iYPos, _oTime):
        """Ensure we have correct row highlighted for the drag"""
        # We lift the idea of changing the drag target from
        # Walter Anger's post to the pygtk mailling list
        # http://www.daa.com.au/pipermail/pygtk/2003-November/006431.html
        oCurPath, _oCol = self.get_cursor()
        oIter = self._oStore.get_iter(oCurPath)
        oFilter = self._oStore.get_value(oIter, 1)
        tCurTarget = self.get_dest_row_at_pos(iXPos, iYPos)
        # Flags to fix row highlghting - we use this method rather then gtk's
        # internal code to workaround
        # https://bugzilla.gnome.org/show_bug.cgi?id=641924 on Windows
        # The logic is: Always set the highlighted row if we're dragging a
        # filter element, but only highlight rows if we can drop there
        # when dragging values
        bSetRow = False
        bFixHighlight = False
        if oDragContext.get_source_widget() is self:
            # We're dragging within the view
            if oFilter is None:
                # Dragging a value
                if self._check_drag_value(oIter, oFilter, tCurTarget,
                        gtk.gdk.ACTION_MOVE, oDragContext):
                    self._fix_highlight(tCurTarget)
                    bSetRow = True
                    bFixHighlight = True
            else:
                # dragging a filter around
                oDragContext.drag_status(gtk.gdk.ACTION_MOVE)
                bSetRow = True
        else:
            # Dragging from outside the view, so different rules
            if oFilter is None or not hasattr(oFilter, 'sBoxType'):
                # Dragging values into the filter (since filter is selected)
                if self._check_drag_value(oIter, oFilter, tCurTarget,
                        gtk.gdk.ACTION_COPY, oDragContext):
                    bSetRow = True
                    bFixHighlight = True
            else:
                # Dragging in a filter element
                oDragContext.drag_status(gtk.gdk.ACTION_COPY)
                bSetRow = True
        if bSetRow:
            if tCurTarget:
                self.set_drag_dest_row(tCurTarget[0], tCurTarget[1])
            else:
                # Set highlight on the root of the tree
                oIter = self._oStore.get_iter_root()
                oDropPath = self._oStore.get_path(oIter)
                self.set_drag_dest_row(oDropPath, gtk.TREE_VIEW_DROP_AFTER)
            if bFixHighlight:
                self._fix_highlight(tCurTarget)
        # Ensure we don't propogate this signal further
        return True

    # pylint: enable-msg=R0913

    def _do_drop_value(self, sData, tRowInfo):
        """Handled dropping in new values"""
        def target_ok(oDropFilter, sFilterName):
            """Check that we have a valid target"""
            if not hasattr(oDropFilter, 'sFilterName'):
                # Can't drop values onto a box filter
                return False
            elif oDropFilter.iValueType not in LIST_TYPES:
                # Not sensible to drop onto entry or null filters
                return False
            elif sFilterName != oDropFilter.sFilterName:
                # Only drag between the same filter types
                return False
            return True

        # values are created as 'NewValue: FilterName[: Section]'
        # We then call back into oValuesWidget to get the actual
        # values, to avoid awkward escaping issues
        oFilterObj, oDropIter = self._oStore.get_drop_filter(tRowInfo)
        sFilterName = sData.split(':', 2)[1].strip()
        if not target_ok(oFilterObj, sFilterName):
            return False
        self._update_values(oFilterObj, oDropIter, sData)

    def _update_values(self, oFilterObj, oIter, sData):
        """Update the values with the given info from the filter"""
        aSelection = self._oValuesWidget.get_selected_values()
        if oFilterObj.iValueType == FilterBoxItem.LIST:
            # Get values from the list
            for sValue in aSelection:
                if sValue not in oFilterObj.aCurValues:
                    oFilterObj.aCurValues.append(sValue)
            oFilterObj.aCurValues.sort()
            self.update_list(oIter, oFilterObj)
        else:
            # List from filter - we need to distinguish between the
            # different parts
            sSection = sData.split(':', 2)[2].strip()
            for sValue in aSelection:
                if sSection == 'Count':
                    if not oFilterObj.aCurValues[0]:
                        oFilterObj.aCurValues[0] = [sValue]
                    elif sValue not in oFilterObj.aCurValues[0]:
                        oFilterObj.aCurValues[0].append(sValue)
                else:
                    if not oFilterObj.aCurValues[1]:
                        oFilterObj.aCurValues[1] = [sValue]
                    elif sValue not in oFilterObj.aCurValues[1]:
                        oFilterObj.aCurValues[1].append(sValue)
            if sSection == 'Count':
                oFilterObj.aCurValues[0].sort()
            else:
                oFilterObj.aCurValues[1].sort()
            self.update_count_list(oIter, oFilterObj)
        return True

    def _do_move_value(self, sData, tRowInfo):
        """Handle moving filter values"""
        def target_ok(oDropFilter, oSourceFilter):
            """Check that we have a valid target"""
            if not hasattr(oDropFilter, 'sFilterName'):
                # Can't drop values onto a box filter
                return False
            elif oDropFilter.iValueType not in LIST_TYPES:
                # Not sensible to drop onto entry or null filters
                return False
            elif oSourceFilter.sFilterName != oDropFilter.sFilterName:
                # Only drag between the same filter types
                return False
            elif oSourceFilter is oDropFilter:
                # Don't drag onto ourselves
                return False
            return True

        sIter = sData.split(':', 1)[1].strip()
        # Check if target point is suitable
        oFilterObj, _oDropIter = self._oStore.get_drop_filter(tRowInfo)
        oIter = self._oStore.get_iter_from_string(sIter)
        oSourceFilter = self._oStore.get_value(
                self._oStore.iter_parent(oIter), 1)
        sValue = unescape_markup(self._oStore.get_value(oIter, 0))
        if not target_ok(oFilterObj, oSourceFilter):
            return False
        aTarget = []
        if oFilterObj.iValueType == FilterBoxItem.LIST:
            aTarget = oFilterObj.aCurValues
        elif oFilterObj.iValueType == FilterBoxItem.LIST_FROM:
            for iIndex in (0, 1):
                # We do this loop because we may need to explicitly set
                # the destination filter to an empty list
                if oSourceFilter.aCurValues[iIndex] and \
                        sValue in oSourceFilter.aCurValues[iIndex]:
                    if not oFilterObj.aCurValues[iIndex]:
                        oFilterObj.aCurValues[iIndex] = []
                    aTarget = oFilterObj.aCurValues[iIndex]
                    break
        if sValue in aTarget:
            # Don't allow duplicate values
            return False
        aTarget.append(sValue)
        aTarget.sort()
        self.remove_value_at_iter(oIter)
        self.load()
        self._oStore.foreach(self._check_for_value, (oFilterObj, sValue))
        return True

    def paste_filter(self, oInsertObj, sData, iIndex):
        """Add a filter to the given filter object"""
        oCurPath, _oCol = self.get_cursor()
        sFilter = [x.strip() for x in sData.split(':', 1)][1]
        tSelInfo = self._oValuesWidget.get_current_pos_and_sel()
        if sFilter == 'Filter Group':
            oInsertObj.add_child_box(oInsertObj.AND)
        else:
            oInsertObj.add_child_item(sFilter)
        # Move to the correct place
        if iIndex >= 0:
            oFilterObj = oInsertObj.pop()
            oInsertObj.insert(iIndex, oFilterObj)
        self.load()
        # Restore selection after load
        self.select_path(oCurPath)
        # Restore values widget selection
        self._oValuesWidget.restore_pos_and_selection(tSelInfo)

    def _do_drop_filter(self, sData, tRowInfo):
        """Handle the moving/dropping of a filter onto the view"""
        # Check we have an acceptable drop position
        oInsertObj, iIndex = self._oStore.get_insert_box(tRowInfo)
        if sData.startswith('NewFilter'):
            self.paste_filter(oInsertObj, sData, iIndex)
        else:
            sIter = [x.strip() for x in sData.split(':', 1)][1]
            # Find the dragged filter and remove it from it's current
            # position
            oSourceIter = self._oStore.get_iter_from_string(sIter)
            oFilterObj = self._oStore.get_value(oSourceIter, 1)
            oParent = self._oStore.get_value(
                    self._oStore.iter_parent(oSourceIter), 1)
            # Check move is legal
            bDoInsert = False
            if oInsertObj is not oFilterObj:
                if not isinstance(oFilterObj, FilterBoxModel) or \
                        not oFilterObj.is_in_model(oInsertObj):
                    bDoInsert = True
            if bDoInsert:
                oParent.remove(oFilterObj)
                oInsertObj.append(oFilterObj)
            else:
                return False
            # Move to the correct place
            if iIndex >= 0:
                oFilterObj = oInsertObj.pop()
                oInsertObj.insert(iIndex, oFilterObj)
            self.load()
            self._oStore.foreach(self._check_for_obj, oFilterObj)
        return True

    def _get_cur_filter(self):
        """Get the currently selected filter path"""
        oCurPath, _oCol = self.get_cursor()
        if oCurPath:
            return self._oStore.get_value(self.oCurSelectIter, 1), oCurPath
        return None, None

    def toggle_negate(self, _oWidget):
        """Toggle the disabled flag for a section of the filter"""
        oFilterObj, _oCurPath = self._get_cur_filter()
        if oFilterObj:
            # This is a a bit inefficient, but allows us to handle updating
            # the value in only one place
            self.set_negate(not oFilterObj.bNegated)

    def set_negate(self, bState):
        """Set the disabled flag for a section of the filter"""
        oFilterObj, _oCurPath = self._get_cur_filter()
        if oFilterObj and oFilterObj.bNegated != bState:
            oFilterObj.bNegated = bState
            self._oStore.fix_filter_name(self.oCurSelectIter, oFilterObj)

    def toggle_disabled(self, _oWidget):
        """Toggle the disabled flag for a section of the filter"""
        oFilterObj, oCurPath = self._get_cur_filter()
        if oFilterObj:
            oFilterObj.bDisabled = not oFilterObj.bDisabled
            self.load(oCurPath)

    def set_disabled(self, bState):
        """Set the disabled flag for a section of the filter"""
        oFilterObj, oCurPath = self._get_cur_filter()
        if oFilterObj and oFilterObj.bDisabled != bState:
            oFilterObj.bDisabled = bState
            # We opt for the lazy approach and reload
            self.load(oCurPath)

    def delete(self, _oIgnore):
        """Delete an filter component from the model

           _oIgnore is so this can be called from the popup menu"""
        oFilterObj, oCurPath = self._get_cur_filter()
        oCurIter = self._oStore.get_iter(oCurPath)
        oCurValue = self._oStore.get_value(oCurIter, 1)
        if oCurValue is oFilterObj:
            # At the filter level
            self.remove_filter_at_iter(self.oCurSelectIter)
        elif oFilterObj.iValueType == FilterBoxItem.NONE:
            # None filter, so delete always removes
            self.remove_filter_at_iter(self.oCurSelectIter)
        elif (oFilterObj.iValueType == FilterBoxItem.LIST or
                oFilterObj.iValueType == FilterBoxItem.ENTRY) \
                        and not oFilterObj.aCurValues:
            # Emptu filter
            self.remove_filter_at_iter(self.oCurSelectIter)
        elif oFilterObj.iValueType == FilterBoxItem.LIST_FROM and \
                not oFilterObj.aCurValues[0] and not oFilterObj.aCurValues[1]:
            # empty list from filter, so remove
            self.remove_filter_at_iter(self.oCurSelectIter)
        else:
            # Deleting a value
            self.remove_value_at_iter(oCurIter)

    def press_button(self, _oWidget, oEvent):
        """Display the popup menu"""
        if oEvent.button == 3:
            iXPos = int(oEvent.x)
            iYPos = int(oEvent.y)
            oTime = oEvent.time
            oPathInfo = self.get_path_at_pos(iXPos, iYPos)
            if oPathInfo is not None:
                oPath, oCol, _iCellX, _iCellY = oPathInfo
                self.grab_focus()
                self.set_cursor(oPath, oCol, False)
                oPopupMenu = BoxModelPopupMenu(self)
                # Show before popup, otherwise menu items aren't drawn properly
                oPopupMenu.show_all()
                oPopupMenu.popup(None, None, None, oEvent.button, oTime)
                return True  # Don't propogate to buttons
        return False

    def drag_element(self, _oBtn, _oContext, oSelectionData, _oInfo, _oTime):
        """Create a drag info for this filter"""
        _oModel, oIter = self.get_selection().get_selected()
        if oIter and self._oStore.iter_depth(oIter) > 0:
            # We don't allow the root node to be dragged
            oFilter = self._oStore.get_value(oIter, 1)
            if oFilter is None:
                # Dragging a value
                sSelect = '%s%s' % (MOVE_VALUE,
                        self._oStore.get_string_from_iter(oIter))
            else:
                # Dragging a filter
                sSelect = '%s%s' % (MOVE_FILTER,
                        self._oStore.get_string_from_iter(oIter))
            oSelectionData.set(oSelectionData.target, 8, sSelect)


class FilterBoxModelEditBox(gtk.VBox):
    """Box to hold the BoxModel view."""
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods

    def __init__(self, oValuesWidget):
        super(FilterBoxModelEditBox, self).__init__()
        self._oValuesWidget = oValuesWidget
        oTreeStore = FilterBoxModelStore()
        self._oTreeView = FilterBoxModelEditView(oTreeStore, oValuesWidget,
                None)
        self.pack_start(self._oTreeView, expand=True)

    def set_box_model(self, oBoxModel):
        """Set the box model to the correct value"""
        self._oTreeView.set_box_model(oBoxModel)



# test_PhysicalCardSetWriter.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Test Writing Card Set handling"""

from sutekh.tests.TestCore import SutekhTest
from sutekh.tests.core.test_PhysicalCardSet import CARD_SET_NAMES, \
        get_phys_cards, make_set_1
from sutekh.core.SutekhObjects import PhysicalCardSet, IPhysicalCardSet
from sutekh.core.CardSetHolder import CardSetWrapper
from sutekh.io.PhysicalCardSetWriter import PhysicalCardSetWriter
from sutekh.io.XmlFileHandling import PhysicalCardSetXmlFile
from StringIO import StringIO
import unittest

EXPECTED_1 = '<physicalcardset author="A test author"'\
        ' name="Test Set 1" sutekh_xml_version="1.3">\n' \
        '  <comment>A test comment</comment>\n' \
        '  <annotations />\n' \
        '  <card count="1" expansion="Jyhad" name=".44 Magnum" />\n' \
        '  <card count="3" expansion="None Specified" name=".44 Magnum" />\n' \
        '  <card count="1" expansion="Lords of the Night"' \
        ' name="AK-47" />\n' \
        '  <card count="1" expansion="None Specified" name="AK-47" />\n' \
        '  <card count="1" expansion="Keepers of Tradition"' \
        ' name="Aaron\'s Feeding Razor" />\n' \
        '  <card count="1" expansion="None Specified" name="Abbot" />\n' \
        '  <card count="1" expansion="Third Edition" name="Abbot" />\n' \
        '  <card count="1" expansion="None Specified" name="Abebe" />\n' \
        '  <card count="1" expansion="Legacy of Blood" name="Abombwe" />\n' \
        '  <card count="1" expansion="None Specified" name="Abombwe" />\n' \
        '  <card count="1" expansion="Promo-20051001"' \
        ' name="Alan Sovereign (Advanced)" />\n' \
        '  <card count="1" expansion="Nights of Reckoning"' \
        ' name="Inez &quot;Nurse216&quot; Villagrande" />\n' \
        '  <card count="1" expansion="Dark Sovereigns"' \
        ' name="Scapelli, The Family &quot;Mechanic&quot;" />\n' \
        '  <card count="1" expansion="Lords of the Night"' \
        ' name="The Path of Blood" />\n' \
        '  <card count="1" expansion="Bloodlines" name="The Siamese" />\n' \
        '</physicalcardset>'

EXPECTED_2 = '<physicalcardset author="A test author" ' \
        'name="Test Set 2" '\
        'sutekh_xml_version="1.3">\n' \
        '  <comment>A formatted test comment\n' \
        'A second line\n' \
        'A third line</comment>\n' \
        '  <annotations>Some Annotations</annotations>\n' \
        '  <card count="1" expansion="Jyhad" name=".44 Magnum" />\n' \
        '  <card count="1" expansion="Lords of the Night" name="AK-47" />\n' \
        '  <card count="1" expansion="Third Edition" name="Abbot" />\n' \
        '  <card count="1" expansion="Legacy of Blood" name="Abombwe" />\n ' \
        ' <card count="1" expansion="Promo-20051001"' \
        ' name="Alan Sovereign (Advanced)" />\n' \
        '</physicalcardset>'

EXPECTED_3 = '<physicalcardset author="" name="Test Set 2" '\
        'sutekh_xml_version="1.3">\n' \
        '  <comment>A formatted test comment\n' \
        'A second line\n' \
        'A third line</comment>\n' \
        '  <annotations>Some Annotations</annotations>\n' \
        '  <card count="1" expansion="Jyhad" name=".44 Magnum" />\n' \
        '  <card count="1" expansion="Lords of the Night" name="AK-47" />\n' \
        '  <card count="1" expansion="Third Edition" name="Abbot" />\n' \
        '  <card count="1" expansion="Legacy of Blood" name="Abombwe" />\n ' \
        ' <card count="1" expansion="Promo-20051001"' \
        ' name="Alan Sovereign (Advanced)" />\n' \
        '</physicalcardset>'


class PhysicalCardSetWriterTests(SutekhTest):
    """class for the Physical Card Set writer tests"""

    def test_physical_card_set_writer(self):
        """Test physical card set writing"""
        # pylint: disable-msg=E1101, R0915, R0914
        # E1101: SQLObject + PyProtocols magic confuses pylint
        # R0915, R0914: Want a long, sequential test case to minimise
        # repeated setups, so it has lots of lines + variables
        aAddedPhysCards = get_phys_cards()
        # We have a physical card list, so create some physical card sets
        oPhysCardSet1 = make_set_1()

        # Check output

        oWriter = PhysicalCardSetWriter()
        oFile = StringIO()
        oWriter.write(oFile, CardSetWrapper(oPhysCardSet1))
        sWriterXML = oFile.getvalue()
        oFile.close()
        self.assertEqual(sWriterXML, EXPECTED_1)
        self.assertEqual(len(sWriterXML), len(EXPECTED_1))

        oPCS = IPhysicalCardSet(CARD_SET_NAMES[0])
        sData = self._round_trip_obj(oWriter, CardSetWrapper(oPCS))
        self.assertEqual(sData, sWriterXML)
        self.assertEqual(len(sData), len(EXPECTED_1))
        self.assertEqual(sData, EXPECTED_1)

        oPhysCardSet2 = PhysicalCardSet(name=CARD_SET_NAMES[1])
        oPhysCardSet2.author = 'A test author'
        oPhysCardSet2.comment = 'A formatted test comment\nA second line\n' \
                'A third line'
        oPhysCardSet2.annotations = 'Some Annotations'

        for iLoop in range(5, 10):
            oPhysCardSet2.addPhysicalCard(aAddedPhysCards[iLoop].id)
            oPhysCardSet2.syncUpdate()

        oFile = StringIO()
        oWriter.write(oFile, CardSetWrapper(oPhysCardSet2))
        sWriterXML = oFile.getvalue()
        oFile.close()
        self.assertEqual(len(sWriterXML), len(EXPECTED_2))
        self.assertEqual(sWriterXML, EXPECTED_2)

        sTempFileName = self._create_tmp_file()
        oFileXML = PhysicalCardSetXmlFile(sTempFileName)
        oFileXML.write(CARD_SET_NAMES[1])
        fIn = open(sTempFileName, 'rU')
        sData = fIn.read()
        self.assertEqual(len(sData), len(EXPECTED_2))
        self.assertEqual(sData, EXPECTED_2)

        # Unset the author
        oPhysCardSet2.author = None
        oFile = StringIO()
        oWriter.write(oFile, CardSetWrapper(oPhysCardSet2))
        sWriterXML = oFile.getvalue()
        oFile.close()
        self.assertEqual(len(sWriterXML), len(EXPECTED_3))
        self.assertEqual(sWriterXML, EXPECTED_3)


if __name__ == "__main__":
    unittest.main()



# CSVImporter.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Simon Cross <hodgestar@gmail.com>,
# GPL - see COPYING for details

"""plugin for managing CSV file imports"""

import gtk
import csv
import gobject
from sutekh.core.SutekhObjects import PhysicalCardSet
from sutekh.io.CSVParser import CSVParser
from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.SutekhDialog import SutekhDialog, do_complaint_error
from sutekh.gui.SutekhFileWidget import SutekhFileButton
from sutekh.gui.GuiCardSetFunctions import import_cs


class CSVImporter(SutekhPlugin):
    """CSV Import plugin.

       Allow the user to select the file, provide information about the
       columns to use and so forth.
       The list of the columns available is updated to reflect the
       currently selected file.
       """
    dTableVersions = {
        PhysicalCardSet: (4, 5, 6),
    }
    aModelsSupported = ("MainWindow",)

    def get_menu_item(self):
        """Overrides method from base class. Register on the 'Import' menu"""
        if not self.check_versions() or not self.check_model_type():
            return None
        oMenuItem = gtk.MenuItem("Import CSV File")
        oMenuItem.connect("activate", self.activate)
        return ('Import Card Set', oMenuItem)

    def activate(self, _oWidget):
        """Handle response form the user"""
        oDlg = self.make_dialog()
        oDlg.run()

    # pylint: disable-msg=W0201
    # defining elements outside of init is OK here, because of plugin structure
    def make_dialog(self):
        """Create the dialog.

           Allow the user to select a file and specify which columns to
           use for card name, card count and expansions.
           """
        self.oDlg = SutekhDialog("Choose CSV File", None,
                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_OK, gtk.RESPONSE_OK,
                 gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

        oLabel = gtk.Label()
        oLabel.set_markup("<b>Select CSV File ...</b>")
        oLabel.set_alignment(0.0, 0.5)
        # pylint: disable-msg=E1101
        # vbox confuses pylint
        self.oDlg.vbox.pack_start(oLabel, padding=5)

        self.oFileChooser = SutekhFileButton(self.parent, "Select a CSV File")
        self.oFileChooser.connect("selection-changed",
                                  self._selected_file_changed)
        self.oDlg.vbox.pack_start(self.oFileChooser)

        self.oDlg.vbox.pack_start(gtk.HSeparator(), padding=5)

        oLabel = gtk.Label()
        oLabel.set_markup("<b>Select columns containing ...</b>")
        oLabel.set_alignment(0.0, 0.5)
        self.oDlg.vbox.pack_start(oLabel)

        oNameBox = gtk.HBox()
        oNameBox.pack_start(gtk.Label("Card name:"))
        self.oCardNameCombo = self._create_column_selector()
        oNameBox.pack_start(self.oCardNameCombo)
        self.oDlg.vbox.pack_start(oNameBox)

        oCountBox = gtk.HBox()
        oCountBox.pack_start(gtk.Label("Card count:"))
        self.oCountCombo = self._create_column_selector()
        oCountBox.pack_start(self.oCountCombo)
        self.oDlg.vbox.pack_start(oCountBox)

        oExpansionBox = gtk.HBox()
        oExpansionBox.pack_start(gtk.Label("Expansion name (optional):"))
        self.oExpansionCombo = self._create_column_selector()
        oExpansionBox.pack_start(self.oExpansionCombo)
        self.oDlg.vbox.pack_start(oExpansionBox)

        self._clear_column_selectors()

        self.oDlg.vbox.pack_start(gtk.HSeparator(), padding=5)

        oLabel = gtk.Label()
        oLabel.set_markup("<b>Import as ...</b>")
        oLabel.set_alignment(0.0, 0.5)
        self.oDlg.vbox.pack_start(oLabel)

        oCardSetNameBox = gtk.HBox()
        oCardSetNameBox.pack_start(gtk.Label("Card Set Name:"))
        self.oSetNameEntry = gtk.Entry()
        oCardSetNameBox.pack_start(self.oSetNameEntry)
        self.oDlg.vbox.pack_start(oCardSetNameBox)

        self.oDlg.connect("response", self.handle_response)
        self.oDlg.set_size_request(400, 400)
        self.oDlg.show_all()

        return self.oDlg

    def _selected_file_changed(self, oFileChooser):
        """Update the column selection options for a new file."""
        sFile = oFileChooser.get_filename()

        # pylint: disable-msg=W0703
        # We want to catch all exceptions here
        try:
            fIn = file(sFile, "rb")
        except Exception:
            self._clear_column_selectors()
            return

        try:
            sLine1 = fIn.next()
            sLine2 = fIn.next()
        except Exception:
            fIn.close()
            self._clear_column_selectors()
            return

        fIn.close()

        try:
            oCsv = csv.reader([sLine1, sLine2])
            aRow = oCsv.next()
            sSample = sLine1 + "\n" + sLine2
            oSniff = csv.Sniffer()
            bHasHeader = oSniff.has_header(sSample)
        except (Exception, csv.Error):
            self._clear_column_selectors()
            return

        if bHasHeader:
            aColumns = [(-1, '-')] + list(enumerate(aRow))
        else:
            aColumns = [(-1, '-')] + [(i, str(i)) for i in range(len(aRow))]

        self.oDlg.set_response_sensitive(gtk.RESPONSE_OK, True)
        self._set_column_selectors(aColumns)

    # pylint: disable-msg=R0201
    # A method for consistency
    def _create_column_selector(self):
        """Create a combo box from which a column can be selected."""
        oListStore = gtk.ListStore(gobject.TYPE_INT, gobject.TYPE_STRING)
        oComboBox = gtk.ComboBox(oListStore)
        oCell = gtk.CellRendererText()
        oComboBox.pack_start(oCell, True)
        oComboBox.add_attribute(oCell, 'text', 1)
        return oComboBox
    # pylint: enable-msg=R0201

    def _clear_column_selectors(self):
        """Clear the column selection lists."""
        # can't click ok until a valid CSV file is chosen
        self.oDlg.set_response_sensitive(gtk.RESPONSE_OK, False)
        self._set_column_selectors([(-1, '-')])

    def _set_column_selectors(self, aColumns):
        """Set the contents of the column selection widgets."""
        for oCombo in (self.oCardNameCombo, self.oCountCombo,
                       self.oExpansionCombo):
            oListStore = oCombo.get_model()
            oListStore.clear()
            for iRow, sHeading in aColumns:
                oIter = oListStore.append(None)
                oListStore.set(oIter, 0, iRow, 1, sHeading)
            oCombo.set_active_iter(oListStore.get_iter_root())

    def handle_response(self, _oWidget, oResponse):
        """Handle the user clicking OK on the dialog.

           Do the actual import.
           """
        if oResponse == gtk.RESPONSE_OK:

            iCardNameColumn, iCountColumn, iExpansionColumn = self._get_cols()

            if iCardNameColumn is None or iCountColumn is None:
                sMsg = "Importing a CSV file requires valid columns for both" \
                        " the card names and card counts."
                do_complaint_error(sMsg)
                self.oDlg.run()
                return

            oParser = CSVParser(iCardNameColumn, iCountColumn,
                                iExpansionColumn, bHasHeader=True)

            sFile = self.oFileChooser.get_filename()

            # pylint: disable-msg=W0703
            # We want to catch all exceptions here
            try:
                fIn = file(sFile, "rb")
            except Exception:
                do_complaint_error("Could not open file '%s'." % sFile)
                self.oDlg.destroy()
                return

            # Tread the well-worn import path
            sSetName = self.oSetNameEntry.get_text().strip()
            import_cs(fIn, oParser, self.parent, sSetName)

        self.oDlg.destroy()

    def _get_cols(self):
        """Get the columns of interest from the user's choices"""
        aCols = []
        for oCombo in (self.oCardNameCombo, self.oCountCombo,
                       self.oExpansionCombo):
            oIter = oCombo.get_active_iter()
            oModel = oCombo.get_model()
            if oIter is None:
                aCols.append(None)
            else:
                iVal = oModel.get_value(oIter, 0)
                if iVal < 0:
                    aCols.append(None)
                else:
                    aCols.append(iVal)
        return aCols


plugin = CSVImporter



# CustomDragIconView.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2010 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""gtk.TreeView class implementing a custom drag icon."""

import gtk


class CustomDragIconView(gtk.TreeView):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """Base class for tree views that fiddle with the drag icon"""

    def __init__(self, oModel):
        self._oModel = oModel
        super(CustomDragIconView, self).__init__(self._oModel)

        # Selecting rows
        self._oSelection = self.get_selection()
        self._aOldSelection = []
        self.bReentrant = False

        # We want to be at the end of the chain, so we can fiddle with the
        # icon after gtk_tree_view sets it
        self.connect_after('drag_begin', self.make_drag_icon)
        self.connect_after('drag_motion', self.drag_motion)

    def make_drag_icon(self, _oWidget, _oDragContext):
        """Drag begin signal handler to set custom icon"""
        # NB: The pygtk docs are wrong (at least for gtk 2.20), as the
        # base gtk_tree_create_row_drag_icon method isn't overridden by
        # subclasses
        # This has to be connected via connect_after as it needs to be
        # called after the stock TreeView drag begin signal so that doesn't
        # overwrite our icon

        iNumSelected = self._oSelection.count_selected_rows()
        # Logic is - if we have multiple rows selected, used the stock
        # DND_MULTIPLE icon, otherwise ensure that the icon reflects the
        # correct text as multiple selection fiddling can cause it to be
        # wrong
        if iNumSelected > 1:
            self.drag_source_set_icon_stock(gtk.STOCK_DND_MULTIPLE)
        elif iNumSelected == 1:
            _oModel, aSelectedRows = self._oSelection.get_selected_rows()
            # Create icon from correct path
            # FIXME: gtk comments suggest this may leak, although docs
            # don't mention this - something to watch for
            oDrawable = self.create_row_drag_icon(aSelectedRows[0])
            self.drag_source_set_icon(oDrawable.get_colormap(), oDrawable)
        # We don't change anything in the nothing selected case

    # pylint: disable-msg=R0201
    # needs to be a method, as children can override this if needed
    def drag_motion(self, _oWidget, oDrag_context, _iXPos, _iYPos,
            _oTimestamp):
        """Set appropriate context during drag + drop."""
        if 'STRING' in oDrag_context.targets:
            oDrag_context.drag_status(gtk.gdk.ACTION_COPY)
            return True
        return False

    # Helper method for implementing custom selection behaviour
    def row_selected(self, oSelection):
        """Change the selection behaviour.

           If we have multiple rows selected, and the user selects
           a single row that is in the selection, we DON'T change
           the selection, but we do update the card text and so on.
           """
        if self.bReentrant:
            # This is here because we alter the selection inside
            # this function (resulting in a nested call).
            # self.bReentrant is set and unset below.
            return None

        # If the selection is empty, clear everything and return
        if oSelection.count_selected_rows() <= 0:
            self._aOldSelection = []
            return None

        _oModel, aList = oSelection.get_selected_rows()
        # Implement the non default selection behaviour.
        tCursorPos = self.get_cursor()
        if len(aList) == 1 and len(self._aOldSelection) > 1 and \
            tCursorPos[0] == aList[0] and aList[0] in self._aOldSelection:
            # reset the list to it's previous state, but set
            # displayed card to this one
            try:
                self.bReentrant = True
                for oPath in self._aOldSelection:
                    oSelection.select_path(oPath)
            finally:
                self.bReentrant = False
            oPath = aList[0]
        else:
            _oModel, aList = oSelection.get_selected_rows()
            if not aList:
                self._aOldSelection = []
                return None

            if len(aList) <= len(self._aOldSelection):
                oPath = aList[-1]
            else:
                # Find the last entry in aList that's not in _aOldSelection
                oPath = [x for x in aList if x not in self._aOldSelection][-1]
            self._aOldSelection = aList
        return oPath



# SutekhObjectCache.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2007 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Cache various objects used by sutekh to speed up database queries."""

from sutekh.core.SutekhObjects import AbstractCard, RarityPair, Rarity, Clan, \
        Discipline, DisciplinePair, CardType, Expansion, Ruling, Sect, Title, \
        Creed, Virtue, PhysicalCard, Keyword, Artist, init_cache


class SutekhObjectCache(object):
    """Holds references to commonly used database objects so that they don't
       get removed from the cache during big reads.

       Including AbstractCard in the cache gives about a 40% speed up on
       filtering at the cost of using about 3MB extra memory.

       Including Ruling costs about an extra 1MB for no real speed up, but
       we threw it in anyway (on the assumption it may be useful sometime
       in the future).
       """

    def __init__(self):
        aTypesToCache = [Rarity, Expansion, RarityPair, Discipline,
                DisciplinePair, Clan, CardType, AbstractCard, Ruling,
                Creed, Virtue, Sect, Title, PhysicalCard, Keyword,
                Artist]

        self._dCache = {}
        for cType in aTypesToCache:
            self._dCache[cType] = list(cType.select())

        init_cache()



# WriteLackeyCCG.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Writer for the Lackey CCG file format.
     (AFAICT, the tabs aren't required, but Lackey uses them, so I'm playing
      safe)

   Example:

   2\tCard 1
   3\tCard 2
   1\tCard 3
   Crypt:
   1\tVampire 1
   2\tVampire 2


   """

import unicodedata
from sutekh.core.ELDBUtilities import type_of_card
from sutekh.core.SutekhObjects import canonical_to_csv, IAbstractCard


def lackey_name(oCard):
    """Escape the card name to Lackey CCG's requirements"""
    sName = oCard.name
    if oCard.level is not None:
        sName = sName.replace("(Advanced)", "Adv.")
    sName = canonical_to_csv(sName)
    # Lackey handles double-quotes a bit oddly, so we must as well
    if oCard.cardtype[0].name == 'Imbued':
        # Lackey only uses '' for Imbued
        sName = sName.replace('"', "''")
    else:
        sName = sName.replace('"', "'")
    sName = unicodedata.normalize('NFKD', sName).encode('ascii', 'ignore')
    return sName


class WriteLackeyCCG(object):
    """Create a string in Lackey CCG format representing a card set."""

    # pylint: disable-msg=R0201
    # Method for consistency
    def _gen_inv(self, oHolder):
        """Process the card set, creating the lines as needed"""
        dCards = {'Crypt': {}, 'Library': {}}
        sResult = ""
        for oCard in oHolder.cards:
            sType = type_of_card(IAbstractCard(oCard))
            sName = lackey_name(IAbstractCard(oCard))
            dCards[sType].setdefault(sName, 0)
            dCards[sType][sName] += 1
        # Sort the output
        # Need to be in this order
        for sType in ['Library', 'Crypt']:
            for sName, iNum in sorted(dCards[sType].items()):
                sResult += '%d\t%s\n' % (iNum, sName)
            if sType == 'Library':
                sResult += 'Crypt:\n'
        return sResult

    # pylint: enable-msg=R0201

    def write(self, fOut, oHolder):
        """Takes file object + card set to write, and writes an Lackey CCG
           deck representing the card set"""
        fOut.write(self._gen_inv(oHolder))



# ARDBTextParser.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# ARDB Text File Deck Parser
# Copyright 2007 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Parser for ARDB text deck format.

   Example deck:

   Deck Name : Followers of Set Preconstructed Deck
   Author : L. Scott Johnson
   Description :
   Followers of Set Preconstructed Starter from Lords of the Night.

   http://www.white-wolf.com/vtes/index.php?line=Checklist_LordsOfTheNight

   Crypt [12 vampires] Capacity min: 2 max: 10 average: 5.84
   ------------------------------------------------------------
   2x Nakhthorheb			  10 OBF PRE SER           Follower :4
   ...

   Library [77 cards]
   ------------------------------------------------------------
   Action [20]
     2x Blithe Acceptance
     4x Dream World
   ...
   """

import re


# State Base Classes
class StateError(Exception):
    """Error in the state transitions."""
    pass


class State(object):
    """Base class for the State Objects."""
    def __init__(self, oHolder):
        self._sData = ""
        self._oHolder = oHolder

    def transition(self, sLine):
        """Transition to next state"""
        raise NotImplementedError

    def data(self, sData):
        """Add data to the state object."""
        self._sData += sData


# State Classes
class NameAndAuthor(State):
    """State for extracting Name and Author."""
    def transition(self, sLine):
        """Process the line for Name and Author - trnaisiotn to Description
           if needed."""
        # Check for crypt line, as description isn't always present
        if sLine.strip().startswith('Crypt ['):
            return Cards(self._oHolder)
        elif sLine.strip().startswith('Crypt: ('):
            return Cards(self._oHolder)
        elif sLine.strip().startswith('Crypt ('):
            return Cards(self._oHolder)

        aParts = sLine.split(':', 1)

        if len(aParts) != 2:
            return self  # Nothing of interest seen

        sKey, sValue = aParts[0].strip(), aParts[1].strip()

        if sKey == "Deck Name":
            self._oHolder.name = sValue
        elif sKey == "Author" or sKey == "Created By" or sKey == "Created by":
            self._oHolder.author = sValue
        elif sKey == "Description":
            oDesc = Description(self._oHolder)
            if len(sValue) > 0:
                oDesc.data(sValue)
            return oDesc

        return self


class Description(State):
    """State for extracting description"""
    def transition(self, sLine):
        """Process the line for the description and transition to Cards
           state if needed."""
        if sLine.strip().startswith('Crypt ['):
            self._oHolder.comment = self._sData
            return Cards(self._oHolder)
        elif sLine.strip().startswith('Crypt: ('):
            self._oHolder.comment = self._sData
            return Cards(self._oHolder)
        elif sLine.strip().startswith('Crypt ('):
            self._oHolder.comment = self._sData
            return Cards(self._oHolder)
        else:
            self.data(sLine)
            return self


class Cards(State):
    """State for extracting the cards"""
    _oCardRe = re.compile(
            r'\s*(?P<cnt>[0-9]+)(\s)*(x)*\s+(?P<name>[^\t\r\n]+)')
    _oAdvRe = re.compile('\sAdv\s')

    def transition(self, sLine):
        """Extract the cards from the data.

           This is the terminating state, so we always return Cards from
           this.
           """
        oMatch = self._oCardRe.match(sLine)
        if oMatch:
            iCnt = int(oMatch.group('cnt'))
            sName = oMatch.group('name').split('  ')[0]
            # We see mixed spaces and tabs in the wild, so we need this
            sName = sName.strip()
            # Check for the advacned string and append advanced if needed
            if self._oAdvRe.search(sLine) and not 'Adv' in sName:
                sName += ' (Advanced)'
            self._oHolder.add(iCnt, sName, None)
        return self


# Parser
class ARDBTextParser(object):
    """Parser for the ARDB Text format."""
    def __init__(self):
        """Create an ARDBTextParser.

           oHolder is a sutekh.core.CardSetHolder.CardSetHolder object (or similar).
           """
        self._oState = None

    def reset(self, oHolder):
        """Reset the parser state"""
        self._oState = NameAndAuthor(oHolder)

    def parse(self, fIn, oHolder):
        """Parse the file"""
        self.reset(oHolder)
        for sLine in fIn:
            self._feed(sLine)

    def _feed(self, sLine):
        """Feed the next line to the current state object, and transition if
           required."""
        self._oState = self._oState.transition(sLine)



# ImportFromZipFile.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details
"""Plugin to import selected card sets from a zip file"""

import gtk
import gobject
import os
from logging import Logger
from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.SutekhDialog import SutekhDialog, do_complaint_error, \
        do_exception_complaint
from sutekh.gui.SutekhFileWidget import ZipFileDialog
from sutekh.gui.ProgressDialog import ProgressDialog, SutekhCountLogHandler
from sutekh.gui.ScrolledList import ScrolledList
from sutekh.io.ZipFileWrapper import ZipFileWrapper
from sutekh.gui.GuiCardSetFunctions import reparent_all_children, \
        update_open_card_sets, get_import_name, PROMPT, RENAME, REPLACE


def _do_rename_parent(sOldName, sNewName, dRemaining):
    """Handle the renaming of a parent card set in the unprocessed list."""
    dResult = {}
    for sName, tInfo in dRemaining.iteritems():
        if tInfo[3] == sOldName:
            dResult[sName] = (tInfo[0], tInfo[1], tInfo[2], sNewName)
        else:
            dResult[sName] = tInfo
    return dResult


def _set_selected_rows(_oButton, oScrolledList, aData):
    """Helper to manage changing the se;ection of the scrolled list"""
    oScrolledList.set_selected_rows(aData)


class ZipFileDirStore(gtk.TreeStore):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """Simple tree store to show card set hierachy in a ScrolledList widget"""
    def __init__(self):
        super(ZipFileDirStore, self).__init__(gobject.TYPE_STRING)

    def fill_list(self, dEscapedList):
        """Fill the list"""
        self.clear()
        aNames = sorted(dEscapedList)
        dAdded = {}
        while aNames:
            for sEntry in aNames[:]:
                sParent = dEscapedList[sEntry][3]
                if sParent:
                    # Need escaped version for comparisons
                    sParent = gobject.markup_escape_text(sParent)
                oIter = None
                if sParent in dAdded:
                    oParIter = dAdded[sParent]
                    oIter = self.append(oParIter)
                elif sParent is None or sParent not in aNames:
                    oIter = self.append(None)
                else:
                    # Can't add this entry yet, so look at next one
                    continue
                self.set(oIter, 0, sEntry)
                dAdded[sEntry] = oIter
                aNames.remove(sEntry)
        self.set_sort_column_id(0, gtk.SORT_ASCENDING)  # Sort the display


class SelectZipFileContents(SutekhDialog):
    """Dialog for querying contents of the zip file"""
    # pylint: disable-msg=R0904
    # gtk.Dialog, so lots of public methods

    def __init__(self, dEscapedList, oParent):
        super(SelectZipFileContents, self).__init__(
                "Select Card Sets to Import", oParent,
                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_OK, gtk.RESPONSE_OK, gtk.STOCK_CANCEL,
                    gtk.RESPONSE_CANCEL))

        self.dEscapedList = dEscapedList

        oModel = ZipFileDirStore()

        # Ask user to select entries to import
        self.oScrolledList = ScrolledList('Available Card Sets', oModel)
        # pylint: disable-msg=E1101
        # vbox confuses pylint
        self.vbox.pack_start(self.oScrolledList)
        self.oScrolledList.set_size_request(450, 300)
        self.oScrolledList.fill_list(self.dEscapedList)
        self.oScrolledList.view.expand_all()
        # Add the various buttons
        # Select all and unselect all
        oSelectAll = gtk.Button('Select All')
        oUnSelectAll = gtk.Button('Unselect All')
        oSelectAll.connect('clicked', _set_selected_rows, self.oScrolledList,
                self.dEscapedList)
        oUnSelectAll.connect('clicked', _set_selected_rows, self.oScrolledList,
                [])
        oSelectButtons = gtk.VBox(False, 2)
        oSelectButtons.pack_start(oSelectAll, expand=False)
        oSelectButtons.pack_start(oUnSelectAll, expand=False)
        self.oPrompt = gtk.RadioButton(None, 'Always Ask', False)
        self.oPrompt.set_active(True)
        self.oReplace = gtk.RadioButton(self.oPrompt,
                'Always replace with new card set', False)
        self.oReplace.set_active(False)
        self.oRename = gtk.RadioButton(self.oPrompt,
                'Always create unique name', False)
        self.oRename.set_active(False)
        oRadioButs = gtk.VBox(False, 2)
        oRadioLabel = gtk.Label()
        oRadioLabel.set_markup('<b>How to handle card set name conflicts?</b>')
        oRadioButs.pack_start(oRadioLabel)
        oRadioButs.pack_start(self.oPrompt, expand=False)
        oRadioButs.pack_start(self.oReplace, expand=False)
        oRadioButs.pack_start(self.oRename, expand=False)

        oButtons = gtk.HBox(False, 2)
        oButtons.pack_start(oSelectButtons, expand=False)
        oButtons.pack_start(oRadioButs)
        self.vbox.pack_start(oButtons, expand=False)

        self.show_all()

    def get_clash_mode(self):
        """Return the selected clash mode"""
        if self.oRename.get_active():
            return RENAME
        elif self.oReplace.get_active():
            return REPLACE
        return PROMPT

    def get_selected(self):
        """Get the list of selected card sets"""
        dSelected = {}
        for sName in self.oScrolledList.get_selection():
            dSelected[sName] = self.dEscapedList[sName]
        return dSelected


class ImportFromZipFile(SutekhPlugin):
    """Extract selected card sets from a zip file."""

    dTableVersions = {}
    aModelsSupported = ("MainWindow",)

    # Dialog and Menu Item Creation

    def get_menu_item(self):
        """Register on the Plugins menu"""
        if not self.check_versions() or not self.check_model_type():
            return None

        oImport = gtk.MenuItem("Import Card Set(s) from zip file")
        oImport.connect("activate", self.make_dialog)
        return ('Import Card Set', oImport)

    # Menu responses

    def make_dialog(self, _oWidget):
        """Create the dialog used to select the zip file"""
        sName = "Select zip file to import from."

        oDlg = ZipFileDialog(self.parent, sName, gtk.FILE_CHOOSER_ACTION_OPEN)
        oDlg.show_all()
        oDlg.run()

        sFilename = oDlg.get_name()
        if sFilename:
            self.handle_response(sFilename)

    def handle_response(self, sFilename):
        """Handle response from the import dialog"""
        if not os.path.exists(sFilename):
            do_complaint_error("Backup file %s does not seem to exist."
                    % sFilename)
            return

        oFile = ZipFileWrapper(sFilename)
        dList = oFile.get_all_entries()
        dEscapedList = {}
        for sName, tInfo in dList.iteritems():
            dEscapedList[self.escape(sName)] = (sName, tInfo[0], tInfo[1],
                    tInfo[2])

        oSelDlg = SelectZipFileContents(dEscapedList, self.parent)
        oResponse = oSelDlg.run()
        dSelected = oSelDlg.get_selected()
        iClashMode = oSelDlg.get_clash_mode()
        oSelDlg.destroy()
        if oResponse == gtk.RESPONSE_OK and len(dSelected) > 0:
            self.do_read_list(oFile, dSelected, iClashMode)

    def do_read_list(self, oFile, dSelected, iClashMode):
        """Read the selected list of card sets"""
        oLogHandler = SutekhCountLogHandler()
        oProgressDialog = ProgressDialog()
        oProgressDialog.set_description("Importing Files")
        oLogger = Logger('Read zip file')
        oLogger.addHandler(oLogHandler)
        oLogHandler.set_dialog(oProgressDialog)
        oLogHandler.set_total(len(dSelected))
        oProgressDialog.show()
        bDone = False
        while not bDone:
            dSelected = self._read_heart(oFile, dSelected, oLogger, iClashMode)
            bDone = len(dSelected) == 0
        oProgressDialog.destroy()

    def _read_heart(self, oFile, dSelected, oLogger, iClashMode):
        """Heart of the reading loop - ensure we read parents before
           children, and correct for renames that occur."""
        # pylint: disable-msg=W0703, R0914
        # @0703: we really do want all the exceptions
        # R0914: Use track a lot of state, so many local variables
        dRemaining = {}
        dRenames = {}
        for sEscapedName, tInfo in dSelected.iteritems():
            sName, sFilename, bParentExists, sParentName = tInfo
            # We aim to ensure that we always keep loaded card sets together,
            # so if the parent name clashes, we want to follow the renamed
            # parent, rather than be added to the existing card set of the
            # same name
            if sParentName is not None and sParentName in dSelected and \
                    sParentName not in dRenames:
                # Do have a parent to look at, so skip for now
                if not bParentExists:
                    # The parent is in the list of cards to read, so it
                    # will exist when we get to it
                    dRemaining[sEscapedName] = (sName, sFilename, True,
                            sParentName)
                else:
                    dRemaining[sEscapedName] = tInfo
                continue
            elif sParentName is not None and sParentName in dRenames:
                # Renamed parent, so adjust this - this will also correctly
                # handle cards that no longer need a parent, since it will
                # set sParentName to None
                sParentName = dRenames[sParentName]
            elif not bParentExists:
                # Parent doesn't exist, but isn't in the list, so turn into
                # top level card set
                sParentName = None
            try:
                oCardSetHolder = oFile.read_single_card_set(sFilename)
                oLogger.info('Read %s' % sName)
                # Ask the user whether to rename, replace or cancel
                oCardSetHolder, aChildren = get_import_name(
                        oCardSetHolder, iClashMode)
                dRenames[sName] = oCardSetHolder.name
                # Check for cardsets that were waiting for us
                dRemaining = _do_rename_parent(sName, oCardSetHolder.name,
                        dRemaining)
                if not oCardSetHolder.name:
                    # We skip this card set
                    continue
                # Ensure we use the right name for the parent
                oCardSetHolder.parent = sParentName
                oCardSetHolder.create_pcs(self.cardlookup)
                reparent_all_children(oCardSetHolder.name, aChildren)
                if self.parent.find_cs_pane_by_set_name(oCardSetHolder.name):
                    # Already open, so update to changes
                    update_open_card_sets(self.parent, oCardSetHolder.name)
                self.reload_pcs_list()
            except Exception, oException:
                sMsg = "Failed to import card set %s.\n\n%s" % (sName,
                        oException)
                do_exception_complaint(sMsg)
        return dRemaining


plugin = ImportFromZipFile



# CardSetManagementModel.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""gtk.TreeModel class the card set list."""

import gtk
import gobject
from sutekh.core.SutekhObjects import PhysicalCardSet, IPhysicalCardSet
from sutekh.core.Filters import NullFilter
from sutekh.gui.ConfigFile import CARDSET_LIST


class CardSetManagementModel(gtk.TreeStore):
    # pylint: disable-msg=R0904
    # gtk.Widget, so lots of public methods
    """TreeModel for the card set list"""
    def __init__(self, oMainWindow):
        # We use 2 columns, one for markup, + one for the name, so we
        # avoid excaping stuff all over the place.
        # This does impose consistency requirements on the Model, but
        # that's all handleded in load
        super(CardSetManagementModel, self).__init__(str, str)
        self._dName2Iter = {}

        self._oMainWin = oMainWindow

        self._bApplyFilter = False  # whether to apply the select filter
        # additional filters for selecting from the list
        self._oSelectFilter = None
        self.oEmptyIter = None

        self._aExcludedSet = set()

    # pylint: disable-msg=W0212, C0103
    # W0212 - we explicitly allow access via these properties
    # C0103 - we allow these names
    applyfilter = property(fget=lambda self: self._bApplyFilter,
            fset=lambda self, x: setattr(self, '_bApplyFilter', x))
    selectfilter = property(fget=lambda self: self._oSelectFilter,
            fset=lambda self, x: setattr(self, '_oSelectFilter', x))

    frame_id = property(fget=lambda self: CARDSET_LIST,
            doc="Frame ID of the card set list (for selecting profiles)")

    # This isn't a card set id, but it's here to support profiles
    # (see CardListModel)
    cardset_id = property(fget=lambda self: CARDSET_LIST,
            doc="Cardset ID of card set list (for selecting profiles)")

    # pylint: enable-msg=W0212, C0103

    # pylint: disable-msg=R0201
    # this should be a method for consistency
    def get_card_set_iterator(self, oFilter):
        """Return an interator over the card set model.

           None may be used to retrieve the entire card set list
           """
        if not oFilter:
            oFilter = NullFilter()
        return oFilter.select(PhysicalCardSet).distinct()
    # pylint: enable-msg=R0201

    def _format_set(self, oSet):
        """Format the card set name for display"""
        sMarkup = gobject.markup_escape_text(oSet.name)
        if oSet.name in self._aExcludedSet:
            sMarkup = '<span foreground="grey">%s</span>' % sMarkup
        elif hasattr(self._oMainWin, 'find_cs_pane_by_set_name') and \
                self._oMainWin.find_cs_pane_by_set_name(oSet.name):
            sMarkup = '<span foreground="blue">%s</span>' % sMarkup
        if oSet.inuse:
            # In use sets are in bold
            sMarkup = '<b>%s</b>' % sMarkup
        return sMarkup

    def exclude_set(self, sSetName):
        """Mark the given set as excluded"""
        self._aExcludedSet.add(sSetName)
        # Update markup if required
        oPath = self.get_path_from_name(sSetName)
        if oPath:
            oIter = self.get_iter(oPath)
            sMarkup = self._format_set(IPhysicalCardSet(sSetName))
            # gtk signals will do the rest for us
            self.set(oIter, 0, sMarkup)

    def unexclude_set(self, sSetName):
        """Unmark the given set as excluded"""
        self._aExcludedSet.discard(sSetName)

    def is_excluded(self, sSetName):
        """Check if this set is excluded"""
        return sSetName in self._aExcludedSet

    def enable_sorting(self):
        """Enable the default sorting behaviour"""
        # This is a method for consistency between the different models
        # see CardListModel.py for more info
        self.set_sort_func(0, self.sort_column)
        # sort on card set name by default, for consistency elsewhere
        self.set_sort_column_id(0, gtk.SORT_ASCENDING)

    def load(self):
        """Load the card sets into the card view"""
        self.clear()
        oCardSetIter = self.get_card_set_iterator(self.get_current_filter())
        self._dName2Iter = {}

        # Disable sorting while we do the insertions - speeds things up
        iSortColumn, iSortOrder = self.get_sort_column_id()
        if iSortColumn is not None:
            self.set_sort_column_id(-2, 0)

        # Loop through the card sets, getting the parent->child relationships
        for oCardSet in oCardSetIter:
            if oCardSet.name in self._dName2Iter:
                # We've already loaded this card set, so skip
                continue
            if oCardSet.parent:
                # Do funky stuff to make sure parent is shown in the view
                oParent = oCardSet
                aToAdd = []
                oIter = None
                while oParent and oParent.name not in self._dName2Iter:
                    aToAdd.insert(0, oParent)  # Insert at the head
                    oParent = oParent.parent
                if oParent and oParent.name in self._dName2Iter:
                    oIter = self._dName2Iter[oParent.name]
            else:
                # Just add
                oIter = None
                aToAdd = [oCardSet]
            for oSet in aToAdd:
                oIter = self.append(oIter)
                sMarkup = self._format_set(oSet)
                self.set(oIter, 0, sMarkup, 1, oSet.name)
                self._dName2Iter[oSet.name] = oIter

        if not self._dName2Iter:
            # Showing nothing
            self.oEmptyIter = self.append(None)
            sText = self._get_empty_text()
            self.set(self.oEmptyIter, 0, sText)

        if iSortColumn is not None:
            self.set_sort_column_id(iSortColumn, iSortOrder)

    def get_current_filter(self):
        """Get the current applied filter."""
        if self.applyfilter:
            return self.selectfilter
        else:
            return None

    def get_name_from_iter(self, oIter):
        """Extract the value at oIter from the model, correcting for encoding
           issues."""
        sCardSetName = self.get_value(oIter, 1)
        if sCardSetName:
            # We can be None when card sets are being deleted.
            sCardSetName = sCardSetName.decode("utf-8")
        return sCardSetName

    def get_path_from_name(self, sName):
        """Get the tree path corresponding to the name"""
        def check_iter(oIter, sName):
            """Recursively descend the children of the given gtk.TreeIter,
               looking for a path matching sName."""
            oPath = None
            if sName == self.get_name_from_iter(oIter):
                return self.get_path(oIter)
            oChildIter = self.iter_children(oIter)
            while oChildIter:
                oPath = check_iter(oChildIter, sName)
                if oPath:
                    return oPath
                oChildIter = self.iter_next(oChildIter)
            return oPath

        oIter = self.get_iter_root()
        while oIter:
            oPath = check_iter(oIter, sName)
            if oPath:
                return oPath
            oIter = self.iter_next(oIter)
        return None

    def get_name_from_path(self, oPath):
        """Get the card set name at oPath."""
        oIter = self.get_iter(oPath)
        return self.get_name_from_iter(oIter)

    def sort_column(self, _oModel, oIter1, oIter2):
        """Custom sort function - ensure that markup doesn't affect sort
           order"""
        oCardSet1 = self.get_name_from_iter(oIter1)
        oCardSet2 = self.get_name_from_iter(oIter2)
        # get_name_from_iter strips markup for us
        return cmp(oCardSet1, oCardSet2)

    def _get_empty_text(self):
        """Get the correct text for an empty model."""
        if self.get_card_set_iterator(None).count() == 0:
            sText = 'Empty'
        else:
            sText = 'No Card Sets found'
        return sText
